"""Native Spark implementations of transform ops.

Mirrors ops.py but operates on native pyspark.sql.DataFrame via
``withColumn`` — never touches pyspark.pandas, so ``_psseries``
corruption is impossible.  Used by ``TransformExecutor._apply_all_distributed``.
"""

from __future__ import annotations

from pyspark.sql import DataFrame as SparkDataFrame
from pyspark.sql import functions as F  # noqa: N812


def spark_impute_null(df: SparkDataFrame, column: str, *, value=0) -> SparkDataFrame:
    if column not in df.columns:
        return df
    if value == "median":
        med = df.agg(F.percentile_approx(F.col(column), 0.5).alias("v")).head()["v"]
        return df.fillna({column: float(med) if med is not None else 0})
    return df.fillna({column: value})


def spark_cap_outlier(df: SparkDataFrame, column: str, *, lower: float = 0, upper: float = 1_000_000) -> SparkDataFrame:
    if column not in df.columns:
        return df
    return df.withColumn(column, F.greatest(F.least(F.col(column), F.lit(upper)), F.lit(lower)))


def spark_type_cast(df: SparkDataFrame, column: str, *, dtype: str = "float") -> SparkDataFrame:
    if column not in df.columns:
        return df
    spark_type = {"float": "double", "int": "int", "string": "string"}.get(dtype, dtype)
    return df.withColumn(column, F.col(column).cast(spark_type))


def spark_drop_column(df: SparkDataFrame, column: str) -> SparkDataFrame:
    return df.drop(column) if column in df.columns else df


def spark_winsorize(df: SparkDataFrame, column: str, *, lower_bound: float = 0, upper_bound: float = 1_000_000) -> SparkDataFrame:
    if column not in df.columns:
        return df
    return df.withColumn(column, F.greatest(F.least(F.col(column), F.lit(upper_bound)), F.lit(lower_bound)))


def spark_segment_aware_cap(df: SparkDataFrame, column: str, *, n_segments: int = 2) -> SparkDataFrame:
    if column not in df.columns:
        return df
    quantiles = df.approxQuantile(column, [0.25, 0.75], 0.01)
    if len(quantiles) == 2:
        q1, q3 = quantiles
        iqr = q3 - q1
        lower, upper = q1 - 1.5 * iqr, q3 + 1.5 * iqr
        return df.withColumn(column, F.greatest(F.least(F.col(column), F.lit(upper)), F.lit(lower)))
    return df


def spark_log_transform(df: SparkDataFrame, column: str) -> SparkDataFrame:
    if column not in df.columns:
        return df
    return df.withColumn(column, F.log1p(F.greatest(F.col(column), F.lit(0)).cast("double")))


def spark_sqrt_transform(df: SparkDataFrame, column: str) -> SparkDataFrame:
    if column not in df.columns:
        return df
    return df.withColumn(column, F.sqrt(F.abs(F.col(column))))


def spark_zero_inflation(df: SparkDataFrame, column: str) -> SparkDataFrame:
    if column not in df.columns:
        return df
    return (
        df.withColumn(f"{column}_is_zero", F.when(F.col(column) == 0, 1).otherwise(0))
          .withColumn(f"{column}_log", F.when(F.col(column) > 0, F.log1p(F.col(column).cast("double"))).otherwise(F.lit(0.0)))
    )


def spark_cap_then_log(df: SparkDataFrame, column: str, *, q99: float | None = None) -> SparkDataFrame:
    if column not in df.columns:
        return df
    if q99 is None:
        quantiles = df.approxQuantile(column, [0.99], 0.01)
        q99 = quantiles[0] if quantiles else None
    if q99 is None:
        return df
    return df.withColumn(column, F.log1p(F.greatest(F.least(F.col(column), F.lit(q99)), F.lit(0)).cast("double")))


def spark_one_hot_encode(df: SparkDataFrame, column: str) -> SparkDataFrame:
    if column not in df.columns:
        return df
    categories = [row[column] for row in df.select(column).distinct().collect() if row[column] is not None]
    for cat in sorted(str(c) for c in categories):
        safe_name = f"{column}_{cat}".replace(" ", "_").replace("-", "_")
        df = df.withColumn(safe_name, F.when(F.col(column) == cat, 1).otherwise(0))
    return df.drop(column)


def spark_feature_select(df: SparkDataFrame, column: str) -> SparkDataFrame:
    return df.drop(column) if column in df.columns else df


def spark_derived_ratio(df: SparkDataFrame, column: str, *, numerator: str, denominator: str) -> SparkDataFrame:
    if numerator not in df.columns or denominator not in df.columns:
        return df
    return df.withColumn(column, F.col(numerator) / F.when(F.col(denominator) == 0, F.lit(None)).otherwise(F.col(denominator)))


def spark_derived_interaction(df: SparkDataFrame, column: str, *, col_a: str, col_b: str) -> SparkDataFrame:
    if col_a not in df.columns or col_b not in df.columns:
        return df
    return df.withColumn(column, F.col(col_a) * F.col(col_b))


def spark_derived_composite(df: SparkDataFrame, column: str, *, columns: list[str]) -> SparkDataFrame:
    valid = [c for c in columns if c in df.columns]
    if not valid:
        return df
    avg_expr = sum(F.coalesce(F.col(c), F.lit(0.0)) for c in valid) / len(valid)
    return df.withColumn(column, avg_expr)


def spark_standard_scale(df: SparkDataFrame, column: str, *, mean: float, scale: float) -> SparkDataFrame:
    if column not in df.columns:
        return df
    return df.withColumn(column, (F.col(column) - mean) / scale)


def spark_minmax_scale(df: SparkDataFrame, column: str, *, scale: float, offset: float) -> SparkDataFrame:
    if column not in df.columns:
        return df
    return df.withColumn(column, F.col(column) * scale + offset)


def spark_yeo_johnson(df: SparkDataFrame, column: str, *, lmbda: float, std_mean: float | None = None, std_scale: float | None = None, standardize: bool = True) -> SparkDataFrame:
    if column not in df.columns:
        return df
    c = F.coalesce(F.col(column), F.lit(0.0))
    if abs(lmbda) < 1e-12:
        pos_expr = F.log1p(c)
    else:
        pos_expr = (F.pow(c + 1, lmbda) - 1) / lmbda
    if abs(lmbda - 2) < 1e-12:
        neg_expr = -F.log1p(-c)
    else:
        neg_expr = -(F.pow(-c + 1, 2 - lmbda) - 1) / (2 - lmbda)
    yj = F.when(c >= 0, pos_expr).otherwise(neg_expr)
    if standardize and std_mean is not None and std_scale is not None:
        yj = (yj - std_mean) / std_scale
    return df.withColumn(column, yj)


# ── Batch ops (single select instead of N withColumn) ─────────────


def spark_batch_log_transform(df: SparkDataFrame, columns: list[str]) -> SparkDataFrame:
    valid = set(columns) & set(df.columns)
    if not valid:
        return df
    return df.select(*[
        F.log1p(F.greatest(F.col(c), F.lit(0)).cast("double")).alias(c) if c in valid else F.col(c)
        for c in df.columns
    ])


def spark_batch_sqrt_transform(df: SparkDataFrame, columns: list[str]) -> SparkDataFrame:
    valid = set(columns) & set(df.columns)
    if not valid:
        return df
    return df.select(*[
        F.sqrt(F.abs(F.col(c))).alias(c) if c in valid else F.col(c)
        for c in df.columns
    ])


def spark_batch_zero_inflation(df: SparkDataFrame, columns: list[str]) -> SparkDataFrame:
    valid = set(columns) & set(df.columns)
    if not valid:
        return df
    existing = [F.col(c) for c in df.columns]
    log_cols = [
        F.when(F.col(c) > 0, F.log1p(F.col(c).cast("double"))).otherwise(F.lit(0.0)).alias(f"{c}_log")
        for c in columns if c in valid
    ]
    flags = [F.when(F.col(c) == 0, 1).otherwise(0).alias(f"{c}_is_zero") for c in columns if c in valid]
    return df.select(*existing, *log_cols, *flags)


def spark_batch_cap_then_log(df: SparkDataFrame, columns_q99: list[tuple[str, float | None]]) -> SparkDataFrame:
    valid_map = {c: q for c, q in columns_q99 if c in df.columns and q is not None}
    if not valid_map:
        return df
    return df.select(*[
        F.log1p(F.greatest(F.least(F.col(c), F.lit(valid_map[c])), F.lit(0)).cast("double")).alias(c)
        if c in valid_map else F.col(c)
        for c in df.columns
    ])


def _spark_yj_expr(col_name: str, lmbda: float, std_mean: float | None, std_scale: float | None, standardize: bool):
    c = F.coalesce(F.col(col_name), F.lit(0.0))
    pos_expr = F.log1p(c) if abs(lmbda) < 1e-12 else (F.pow(c + 1, lmbda) - 1) / lmbda
    neg_expr = -F.log1p(-c) if abs(lmbda - 2) < 1e-12 else -(F.pow(-c + 1, 2 - lmbda) - 1) / (2 - lmbda)
    yj = F.when(c >= 0, pos_expr).otherwise(neg_expr)
    if standardize and std_mean is not None and std_scale is not None:
        yj = (yj - std_mean) / std_scale
    return yj.alias(col_name)


def spark_batch_yeo_johnson(df: SparkDataFrame, params_list: list[tuple[str, float, float | None, float | None, bool]]) -> SparkDataFrame:
    param_map = {col: (lmbda, mean, scale, std) for col, lmbda, mean, scale, std in params_list if col in df.columns}
    if not param_map:
        return df
    return df.select(*[
        _spark_yj_expr(c, *param_map[c]) if c in param_map else F.col(c)
        for c in df.columns
    ])



from typing import List

import nbformat

from ..base import NotebookStage
from .base_stage import StageGenerator


class ModelTrainingStage(StageGenerator):
    @property
    def stage(self) -> NotebookStage:
        return NotebookStage.MODEL_TRAINING

    @property
    def title(self) -> str:
        return "07 - Model Training"

    @property
    def description(self) -> str:
        return "Train baseline models with MLflow experiment tracking."

    def generate_local_cells(self) -> List[nbformat.NotebookNode]:
        target = self.get_target_column()
        test_size = self.config.test_size
        exp_name = self.config.mlflow.experiment_name
        tracking_uri = self.config.mlflow.tracking_uri
        return self.header_cells() + [
            self.cb.section("Imports"),
            self.cb.from_imports_cell({
                "customer_retention.stages.modeling": ["BaselineTrainer", "ModelEvaluator", "DataSplitter"],
                "customer_retention.integrations.adapters": ["get_mlflow"],
                "customer_retention.analysis.visualization": ["ChartBuilder"],
                "customer_retention.stages.temporal": ["SnapshotManager"],
                "customer_retention.analysis.diagnostics": ["LeakageDetector"],
                "pathlib": ["Path"],
                "pandas": ["pd"],
            }),
            self.cb.section("Load Training Snapshot"),
            self.cb.markdown('''**Important**: We load from a versioned snapshot to ensure reproducibility and prevent data leakage.'''),
            self.cb.code('''snapshot_manager = SnapshotManager(Path("./experiments/data"))
latest_snapshot = snapshot_manager.get_latest_snapshot()

if latest_snapshot:
    df, snapshot_metadata = snapshot_manager.load_snapshot(latest_snapshot)
    print(f"Loaded snapshot: {latest_snapshot}")
    print(f"Snapshot cutoff date: {snapshot_metadata.cutoff_date}")
    print(f"Data hash: {snapshot_metadata.data_hash}")
    print(f"Rows: {snapshot_metadata.row_count}")
else:
    from customer_retention.integrations.adapters.factory import get_delta
    storage = get_delta()
    df = storage.read("./experiments/data/gold/customers_selected")
    snapshot_metadata = None
    print(f"Warning: No snapshot found, loading from gold layer: {df.shape}")'''),
            self.cb.section("Prepare Train/Test Split"),
            self.cb.code(f'''target_col = "target" if "target" in df.columns else "{target}"
id_cols = ["entity_id"] if "entity_id" in df.columns else {self.get_identifier_columns()}
temporal_cols = ["feature_timestamp", "label_timestamp", "label_available_flag"]
exclude_cols = id_cols + [target_col] + temporal_cols

feature_cols = [c for c in df.columns if c not in exclude_cols]
print(f"Using {{len(feature_cols)}} features (excluded: {{exclude_cols}})")

X = df[feature_cols]
y = df[target_col]

splitter = DataSplitter(test_size={test_size}, stratify=True, random_state=42)
X_train, X_test, y_train, y_test = splitter.split(X, y)
print(f"Train: {{len(X_train)}}, Test: {{len(X_test)}}")'''),
            self.cb.section("Run Leakage Detection"),
            self.cb.code('''detector = LeakageDetector()
leakage_result = detector.run_all_checks(X_train, y_train)

if not leakage_result.passed:
    print("WARNING: Leakage detected!")
    for issue in leakage_result.critical_issues:
        print(f"  CRITICAL: {issue.feature} - {issue.recommendation}")
else:
    print("Leakage check PASSED")'''),
            self.cb.section("Setup MLflow Tracking"),
            self.cb.code(f'''mlflow_adapter = get_mlflow(tracking_uri="{tracking_uri}", force_local=True)
experiment_name = "{exp_name}"
print(f"MLflow tracking URI: {tracking_uri}")

snapshot_params = {{}}
if snapshot_metadata:
    snapshot_params = {{
        "snapshot_id": snapshot_metadata.snapshot_id,
        "snapshot_version": snapshot_metadata.version,
        "snapshot_cutoff": str(snapshot_metadata.cutoff_date),
        "snapshot_hash": snapshot_metadata.data_hash,
    }}'''),
            self.cb.section("Train Baseline Models"),
            self.cb.code('''from sklearn.linear_model import LogisticRegression
from sklearn.ensemble import RandomForestClassifier, GradientBoostingClassifier

models = {
    "logistic_regression": LogisticRegression(class_weight="balanced", max_iter=1000),
    "random_forest": RandomForestClassifier(class_weight="balanced", n_estimators=100, random_state=42),
    "gradient_boosting": GradientBoostingClassifier(n_estimators=100, random_state=42),
}

results = {}
for name, model in models.items():
    mlflow_adapter.start_run(experiment_name, run_name=name)
    model.fit(X_train, y_train)
    y_pred = model.predict(X_test)
    y_prob = model.predict_proba(X_test)[:, 1] if hasattr(model, "predict_proba") else y_pred

    evaluator = ModelEvaluator()
    metrics = evaluator.evaluate(y_test, y_pred, y_prob)
    results[name] = {"model": model, "metrics": metrics, "y_pred": y_pred, "y_prob": y_prob}

    all_params = {**model.get_params(), **snapshot_params}
    mlflow_adapter.log_params(all_params)
    mlflow_adapter.log_metrics(metrics)
    mlflow_adapter.log_model(model, "model")
    mlflow_adapter.end_run()
    print(f"{name}: AUC={metrics.get('roc_auc', 0):.4f}, F1={metrics.get('f1', 0):.4f}")'''),
            self.cb.section("Compare Models"),
            self.cb.code('''charts = ChartBuilder()
fig = charts.model_comparison_grid(results, y_test)
fig.show()'''),
            self.cb.section("Save Best Model"),
            self.cb.code('''best_model_name = max(results, key=lambda k: results[k]["metrics"].get("roc_auc", 0))
best_model = results[best_model_name]["model"]
import joblib
joblib.dump(best_model, "./experiments/data/models/best_model.joblib")
print(f"Best model: {best_model_name}")'''),
        ]

    def generate_databricks_cells(self) -> List[nbformat.NotebookNode]:
        catalog = self.config.feature_store.catalog
        schema = self.config.feature_store.schema
        target = self.get_target_column()
        exp_name = self.config.mlflow.experiment_name
        model_name = self.config.mlflow.model_name
        return self.header_cells() + [
            self.cb.section("Load Selected Features"),
            self.cb.code(f'''df = spark.table("{catalog}.{schema}.gold_selected")'''),
            self.cb.section("Prepare Features Vector"),
            self.cb.code(f'''from pyspark.ml.feature import VectorAssembler

target_col = "{target}"
feature_cols = [c for c in df.columns if c not in {self.get_identifier_columns()} + [target_col]]

assembler = VectorAssembler(inputCols=feature_cols, outputCol="features", handleInvalid="skip")
df_ml = assembler.transform(df).select("features", target_col)
train_df, test_df = df_ml.randomSplit([0.8, 0.2], seed=42)
print(f"Train: {{train_df.count()}}, Test: {{test_df.count()}}")'''),
            self.cb.section("Setup MLflow"),
            self.cb.code(f'''import mlflow
from customer_retention.core.config.experiments import get_mlflow_dfs_tmpdir
mlflow.set_experiment("/Users/{{spark.conf.get('spark.databricks.notebook.username', 'default')}}/{exp_name}")
_dfs_tmpdir = get_mlflow_dfs_tmpdir()'''),
            self.cb.section("Train Gradient Boosted Trees"),
            self.cb.code(f'''from pyspark.ml.classification import GBTClassifier
from pyspark.ml.evaluation import BinaryClassificationEvaluator

with mlflow.start_run(run_name="gbt_baseline"):
    gbt = GBTClassifier(featuresCol="features", labelCol="{target}", maxIter=100)
    model = gbt.fit(train_df)

    predictions = model.transform(test_df)
    evaluator = BinaryClassificationEvaluator(labelCol="{target}", metricName="areaUnderROC")
    auc = evaluator.evaluate(predictions)

    mlflow.log_param("maxIter", 100)
    mlflow.log_metric("auc_roc", auc)
    mlflow.spark.log_model(model, "model", dfs_tmpdir=_dfs_tmpdir)

    run_id = mlflow.active_run().info.run_id
    print(f"AUC: {{auc:.4f}}, Run ID: {{run_id}}")'''),
            self.cb.section("Register Model"),
            self.cb.code(f'''model_uri = f"runs:/{{run_id}}/model"
mlflow.register_model(model_uri, "{catalog}.{schema}.{model_name}")
print(f"Model registered: {catalog}.{schema}.{model_name}")'''),
        ]



#!/usr/bin/env python3
import argparse
import sys
from pathlib import Path
from typing import Dict, List, Optional, Tuple

import nbformat

from customer_retention.generators.notebook_sync.cell_id_standardizer import (
    split_blended_cell,
    standardize_cell_ids,
)
from customer_retention.generators.notebook_sync.cell_types import CellSyncType, prepend_magic_comment

SplitSpec = Tuple[str, int]

SPLIT_MAP: Dict[str, List[SplitSpec]] = {
    "00_start_here.ipynb": [
        ("cell-1", 6),
        ("cell-5", 4),
        ("cell-11", 4),
        ("sidz4bufnc", 27),
        ("6vilcr3g1hr", 4),
    ],
    "01_data_discovery.ipynb": [
        ("cell-3", 60),
    ],
    "01b_temporal_quality.ipynb": [
        ("cell-3", 1),
        ("cell-5", 3),
    ],
    "01c_temporal_patterns.ipynb": [
        ("19b4e63f", 1),
        ("9e9c32ab", 4),
        ("0e8069c1", 7),
    ],
    "01d_event_aggregation.ipynb": [
        ("cell-3", 1),
        ("cell-7", 5),
    ],
    "10_spec_generation.ipynb": [
        ("b805976d", 35),
    ],
}

TAG_ONLY_MAP: Dict[str, Dict[str, CellSyncType]] = {
    "01_data_discovery.ipynb": {
        "active-dataset-code": CellSyncType.USER_CODE,
    },
    "09_business_alignment.ipynb": {
        "a947050e": CellSyncType.USER_CODE,
        "aec679b2": CellSyncType.USER_CODE,
        "27e74568": CellSyncType.USER_CODE,
        "f6682f6e": CellSyncType.USER_CODE,
        "97c21062": CellSyncType.USER_CODE,
    },
}

DUMMY_SPLIT_MAP: Dict[str, List[SplitSpec]] = {
    "99_dummy_migration.ipynb": [
        ("cell-1", 4),
    ],
}

DUMMY_TAG_ONLY_MAP: Dict[str, Dict[str, CellSyncType]] = {
    "99_dummy_migration.ipynb": {
        "cell-2": CellSyncType.CONFIG,
        "active-dataset-code": CellSyncType.USER_CODE,
    },
}


def _ensure_source_lines(cell) -> List[str]:
    src = cell.source
    if isinstance(src, str):
        return src.splitlines(keepends=True)
    return list(src)



def _find_cell_index_optional(nb: nbformat.NotebookNode, cell_id: str) -> Optional[int]:
    for i, cell in enumerate(nb.cells):
        if cell.id == cell_id:
            return i
    return None


def apply_splits(
    nb: nbformat.NotebookNode,
    filename: str,
    split_map: Dict[str, List[SplitSpec]],
) -> int:
    splits = split_map.get(filename, [])
    offset = 0
    for cell_id, split_after_line in splits:
        idx = _find_cell_index_optional(nb, cell_id)
        if idx is None:
            continue
        cell = nb.cells[idx]
        cell.source = _ensure_source_lines(cell)
        config_cell, code_cell = split_blended_cell(
            cell, split_after_line, f"{cell_id}-cfg", f"{cell_id}-code"
        )
        nb.cells[idx : idx + 1] = [config_cell, code_cell]
        offset += 1
    return offset


def apply_tags(
    nb: nbformat.NotebookNode,
    filename: str,
    tag_only_map: Dict[str, Dict[str, CellSyncType]],
) -> None:
    tags = tag_only_map.get(filename, {})
    for cell_id, cell_type in tags.items():
        idx = _find_cell_index_optional(nb, cell_id)
        if idx is None:
            continue
        cell = nb.cells[idx]
        lines = _ensure_source_lines(cell)
        cell.source = prepend_magic_comment(lines, cell_type)


def apply_config_tags_to_split_cells(nb: nbformat.NotebookNode, filename: str, split_map: Dict[str, List[SplitSpec]]) -> None:
    splits = split_map.get(filename, [])
    for cell_id, _ in splits:
        cfg_id = f"{cell_id}-cfg"
        idx = _find_cell_index_optional(nb, cfg_id)
        if idx is None:
            continue
        cell = nb.cells[idx]
        lines = _ensure_source_lines(cell)
        cell.source = prepend_magic_comment(lines, CellSyncType.CONFIG)


def migrate_notebook(
    nb_path: Path,
    dry_run: bool = False,
    split_map: Optional[Dict[str, List[SplitSpec]]] = None,
    tag_only_map: Optional[Dict[str, Dict[str, CellSyncType]]] = None,
) -> Tuple[int, int]:
    if split_map is None:
        split_map = SPLIT_MAP
    if tag_only_map is None:
        tag_only_map = TAG_ONLY_MAP

    filename = nb_path.name
    nb = nbformat.read(str(nb_path), as_version=4)
    original_count = len(nb.cells)

    apply_splits(nb, filename, split_map)
    apply_config_tags_to_split_cells(nb, filename, split_map)
    apply_tags(nb, filename, tag_only_map)
    nb = standardize_cell_ids(nb, filename)

    nbformat.validate(nb)

    if not dry_run:
        nbformat.write(nb, str(nb_path))

    return original_count, len(nb.cells)


def migrate_all(
    notebooks_dir: Path,
    dry_run: bool = False,
    split_map: Optional[Dict[str, List[SplitSpec]]] = None,
    tag_only_map: Optional[Dict[str, Dict[str, CellSyncType]]] = None,
) -> List[str]:
    notebooks = sorted(notebooks_dir.glob("*.ipynb"))
    if not notebooks:
        print(f"No notebooks found in {notebooks_dir}")
        return []

    print(f"Found {len(notebooks)} notebooks in {notebooks_dir}")
    migrated = []
    for nb_path in notebooks:
        try:
            orig, final = migrate_notebook(nb_path, dry_run=dry_run, split_map=split_map, tag_only_map=tag_only_map)
            label = "would migrate" if dry_run else "migrated"
            splits = final - orig
            suffix = f" ({splits} splits)" if splits else ""
            print(f"  {nb_path.name} — {label}{suffix}")
            migrated.append(nb_path.name)
        except Exception as e:
            print(f"  {nb_path.name} — ERROR: {e}")
    return migrated


def main():
    parser = argparse.ArgumentParser(
        description="Migrate notebook cell IDs to standardized format",
    )
    parser.add_argument(
        "--notebooks-dir",
        default="exploration_notebooks",
        help="Directory containing notebooks (default: exploration_notebooks)",
    )
    parser.add_argument(
        "--dry-run",
        action="store_true",
        help="Show what would change without modifying files",
    )
    args = parser.parse_args()
    notebooks_dir = Path(args.notebooks_dir).resolve()
    if not notebooks_dir.exists():
        print(f"Directory not found: {notebooks_dir}")
        sys.exit(1)
    migrate_all(notebooks_dir, dry_run=args.dry_run)


if __name__ == "__main__":
    main()



"""
Exploration Manager for managing multiple dataset explorations.

Provides functionality for:
- Discovering and loading exploration findings
- Managing dataset inclusion/exclusion
- Detecting relationships between datasets
- Planning aggregations for multi-dataset analysis
"""
from __future__ import annotations

from dataclasses import dataclass, field
from pathlib import Path
from typing import TYPE_CHECKING, Any, Dict, List, Optional

import yaml

from customer_retention.core.config.column_config import DatasetGranularity

from .findings import ExplorationFindings

if TYPE_CHECKING:
    from customer_retention.generators.pipeline_generator.models import FeatureExclusion

    from .layered_recommendations import RecommendationRegistry


@dataclass
class DatasetInfo:
    """Information about a discovered dataset."""

    name: str
    findings_path: str
    source_path: str
    granularity: DatasetGranularity
    row_count: int
    column_count: int
    entity_column: Optional[str] = None
    time_column: Optional[str] = None
    target_column: Optional[str] = None
    raw_source_path: Optional[str] = None
    excluded: bool = False
    feature_exclusions: List[FeatureExclusion] = field(default_factory=list)
    excluded_leaking_features: List[str] = field(default_factory=list)


@dataclass
class DatasetRelationshipInfo:
    """Information about relationship between two datasets."""

    left_dataset: str
    right_dataset: str
    left_columns: list
    right_columns: list
    relationship_type: str  # one_to_one, one_to_many, many_to_many
    confidence: float = 1.0
    auto_detected: bool = False

    @property
    def left_column(self) -> str:
        return self.left_columns[0] if self.left_columns else ""

    @property
    def right_column(self) -> str:
        return self.right_columns[0] if self.right_columns else ""


@dataclass
class AggregationPlanItem:
    """Plan for aggregating one event dataset."""

    dataset_name: str
    entity_column: str
    time_column: str
    windows: List[str]
    value_columns: List[str]
    agg_funcs: List[str]


@dataclass
class MultiDatasetFindings:
    """Findings for multiple related datasets."""

    datasets: Dict[str, DatasetInfo] = field(default_factory=dict)
    relationships: List[DatasetRelationshipInfo] = field(default_factory=list)
    primary_entity_dataset: Optional[str] = None
    event_datasets: List[str] = field(default_factory=list)
    excluded_datasets: List[str] = field(default_factory=list)
    aggregation_windows: List[str] = field(
        default_factory=lambda: ["24h", "7d", "30d", "90d", "180d", "365d", "all_time"]
    )
    composite_dataset_name: Optional[str] = None
    notes: Dict[str, Any] = field(default_factory=dict)

    @property
    def selected_datasets(self) -> Dict[str, DatasetInfo]:
        """Return only datasets that are not excluded."""
        return {
            name: info
            for name, info in self.datasets.items()
            if name not in self.excluded_datasets and not info.excluded
        }

    def exclude_dataset(self, name: str) -> None:
        """Exclude a dataset from the pipeline."""
        if name in self.datasets:
            if name not in self.excluded_datasets:
                self.excluded_datasets.append(name)
            self.datasets[name].excluded = True

    def select_dataset(self, name: str) -> None:
        """Re-include a previously excluded dataset."""
        if name in self.excluded_datasets:
            self.excluded_datasets.remove(name)
        if name in self.datasets:
            self.datasets[name].excluded = False

    def add_relationship(
        self,
        left_dataset: str,
        right_dataset: str,
        left_column: str,
        right_column: str,
        relationship_type: str = "one_to_many",
        confidence: float = 1.0,
    ) -> None:
        """Add a relationship between datasets."""
        rel = DatasetRelationshipInfo(
            left_dataset=left_dataset,
            right_dataset=right_dataset,
            left_columns=[left_column],
            right_columns=[right_column],
            relationship_type=relationship_type,
            confidence=confidence,
            auto_detected=False,
        )
        self.relationships.append(rel)

    def get_aggregation_plan(self) -> Dict[str, AggregationPlanItem]:
        """Generate aggregation plan for all event datasets."""
        plan = {}

        for dataset_name in self.event_datasets:
            if dataset_name in self.excluded_datasets:
                continue

            dataset_info = self.datasets.get(dataset_name)
            if dataset_info and dataset_info.entity_column and dataset_info.time_column:
                plan[dataset_name] = AggregationPlanItem(
                    dataset_name=dataset_name,
                    entity_column=dataset_info.entity_column,
                    time_column=dataset_info.time_column,
                    windows=self.aggregation_windows.copy(),
                    value_columns=[],  # To be filled by user
                    agg_funcs=["sum", "mean", "count"],
                )

        return plan

    def to_recommendation_registry(self) -> "RecommendationRegistry":
        """Create a RecommendationRegistry from selected datasets."""
        from .layered_recommendations import RecommendationRegistry

        registry = RecommendationRegistry()

        for name, info in self.selected_datasets.items():
            registry.add_source(name, info.source_path)

        if self.primary_entity_dataset and self.primary_entity_dataset in self.selected_datasets:
            primary_info = self.datasets[self.primary_entity_dataset]
            entity_col = primary_info.entity_column or "id"
            time_col = primary_info.time_column
            registry.init_silver(entity_col, time_col)

            if primary_info.target_column:
                registry.init_gold(primary_info.target_column)

        for rel in self.relationships:
            if (
                rel.left_dataset in self.selected_datasets
                and rel.right_dataset in self.selected_datasets
                and registry.silver
            ):
                registry.add_silver_join(
                    rel.left_dataset,
                    rel.right_dataset,
                    [rel.left_column],
                    rel.relationship_type,
                    f"Join {rel.left_dataset} with {rel.right_dataset}",
                )

        return registry

    def save(self, path) -> None:
        """Save multi-dataset findings to YAML."""
        data = {
            "datasets": {
                name: {
                    "name": info.name,
                    "findings_path": info.findings_path,
                    "source_path": info.source_path,
                    "granularity": info.granularity.value,
                    "row_count": info.row_count,
                    "column_count": info.column_count,
                    "entity_column": info.entity_column,
                    "time_column": info.time_column,
                    "target_column": info.target_column,
                    "raw_source_path": info.raw_source_path,
                    "excluded": info.excluded,
                    "feature_exclusions": [
                        {
                            "column": e.column,
                            "blocked_categories": e.blocked_categories,
                            "blocked_funcs": e.blocked_funcs,
                            "rationale": e.rationale,
                        }
                        for e in info.feature_exclusions
                    ],
                    "excluded_leaking_features": info.excluded_leaking_features,
                }
                for name, info in self.datasets.items()
            },
            "relationships": [
                {
                    "left_dataset": rel.left_dataset,
                    "right_dataset": rel.right_dataset,
                    "left_columns": rel.left_columns,
                    "right_columns": rel.right_columns,
                    "relationship_type": rel.relationship_type,
                    "confidence": rel.confidence,
                    "auto_detected": rel.auto_detected,
                }
                for rel in self.relationships
            ],
            "primary_entity_dataset": self.primary_entity_dataset,
            "event_datasets": self.event_datasets,
            "excluded_datasets": self.excluded_datasets,
            "aggregation_windows": self.aggregation_windows,
            "composite_dataset_name": self.composite_dataset_name,
            "notes": self.notes,
        }

        p = path if isinstance(path, Path) else Path(str(path))
        p.parent.mkdir(parents=True, exist_ok=True)
        with p.open("w") as f:
            yaml.dump(data, f, default_flow_style=False, sort_keys=False)

    @classmethod
    def load(cls, path) -> MultiDatasetFindings:
        """Load multi-dataset findings from YAML."""
        from customer_retention.generators.pipeline_generator.models import FeatureExclusion

        p = path if isinstance(path, Path) else Path(str(path))
        with p.open("r") as f:
            data = yaml.safe_load(f)

        datasets = {}
        for name, info in data.get("datasets", {}).items():
            datasets[name] = DatasetInfo(
                name=info["name"],
                findings_path=info["findings_path"],
                source_path=info["source_path"],
                granularity=DatasetGranularity(info["granularity"]),
                row_count=info["row_count"],
                column_count=info["column_count"],
                entity_column=info.get("entity_column"),
                time_column=info.get("time_column"),
                target_column=info.get("target_column"),
                raw_source_path=info.get("raw_source_path"),
                excluded=info.get("excluded", False),
                feature_exclusions=[
                    FeatureExclusion(
                        column=e["column"],
                        blocked_categories=e.get("blocked_categories", []),
                        blocked_funcs=e.get("blocked_funcs", []),
                        rationale=e.get("rationale", ""),
                    )
                    for e in info.get("feature_exclusions", [])
                ],
                excluded_leaking_features=info.get("excluded_leaking_features", []),
            )

        relationships = [
            DatasetRelationshipInfo(
                left_dataset=rel["left_dataset"],
                right_dataset=rel["right_dataset"],
                left_columns=rel.get("left_columns") or [rel["left_column"]],
                right_columns=rel.get("right_columns") or [rel["right_column"]],
                relationship_type=rel["relationship_type"],
                confidence=rel.get("confidence", 1.0),
                auto_detected=rel.get("auto_detected", False),
            )
            for rel in data.get("relationships", [])
        ]

        return cls(
            datasets=datasets,
            relationships=relationships,
            primary_entity_dataset=data.get("primary_entity_dataset"),
            event_datasets=data.get("event_datasets", []),
            excluded_datasets=data.get("excluded_datasets", []),
            aggregation_windows=data.get(
                "aggregation_windows", ["24h", "7d", "30d", "90d", "180d", "365d", "all_time"]
            ),
            composite_dataset_name=data.get("composite_dataset_name"),
            notes=data.get("notes", {}),
        )


class ExplorationManager:
    """Manages multiple exploration findings."""

    def __init__(self, explorations_dir, findings_paths: Optional[List[Path]] = None):
        self.explorations_dir = explorations_dir if hasattr(explorations_dir, "glob") else Path(explorations_dir)
        self._findings_paths = findings_paths
        self._findings_cache: Dict[str, ExplorationFindings] = {}
        self._excluded_datasets: set = set()

    def discover_findings(self, prefer_aggregated: bool = True) -> List[Path]:
        """Discover all findings files in the explorations directory.

        Excludes multi_dataset_findings.yaml as it has a different structure.

        Args:
            prefer_aggregated: If True, when both event-level and aggregated findings
                exist for the same dataset, only return the aggregated one.
        """
        if self._findings_paths is not None:
            return self._findings_paths

        if not self.explorations_dir.exists():
            return []

        all_files = [f for f in self.explorations_dir.glob("*_findings.yaml") if "multi_dataset" not in f.name]

        if not prefer_aggregated:
            return all_files

        return self._filter_prefer_aggregated(all_files)

    def _filter_prefer_aggregated(self, files: List[Path]) -> List[Path]:
        """Filter findings files to prefer aggregated over event-level.

        Groups files by base dataset name and returns aggregated version when available.
        """
        aggregated = {f for f in files if "_aggregated" in f.name}
        non_aggregated = [f for f in files if "_aggregated" not in f.name]

        if not aggregated:
            return files

        result = list(aggregated)
        aggregated_base_names = {self._get_base_name(f) for f in aggregated}

        for f in non_aggregated:
            base_name = self._get_base_name(f)
            if base_name not in aggregated_base_names:
                result.append(f)

        return result

    def _get_base_name(self, path: Path) -> str:
        """Extract base dataset name without aggregated suffix."""
        stem = path.stem.replace("_findings", "")
        if "_aggregated" in stem:
            stem = stem.rsplit("_aggregated", 1)[0]
        return stem

    def get_skipped_event_findings(self) -> List[Path]:
        """Return event-level findings that were skipped in favor of aggregated versions."""
        if not self.explorations_dir.exists():
            return []

        all_files = [f for f in self.explorations_dir.glob("*_findings.yaml") if "multi_dataset" not in f.name]

        aggregated = {f for f in all_files if "_aggregated" in f.name}
        if not aggregated:
            return []

        aggregated_base_names = {self._get_base_name(f) for f in aggregated}
        return [f for f in all_files if "_aggregated" not in f.name and self._get_base_name(f) in aggregated_base_names]

    def load_findings(self, name_pattern: str) -> Optional[ExplorationFindings]:
        """Load findings by name pattern (partial match)."""
        for path in self.discover_findings():
            if name_pattern.lower() in path.stem.lower():
                if str(path) not in self._findings_cache:
                    self._findings_cache[str(path)] = ExplorationFindings.load(str(path))
                return self._findings_cache[str(path)]
        return None

    def list_datasets(self, include_excluded: bool = False) -> List[DatasetInfo]:
        """List all discovered datasets with their info."""
        datasets = []

        for path in self.discover_findings():
            findings = ExplorationFindings.load(str(path))

            # Determine granularity
            if findings.is_time_series and findings.time_series_metadata:
                granularity = findings.time_series_metadata.granularity
                entity_col = findings.time_series_metadata.entity_column
                time_col = findings.time_series_metadata.time_column
            else:
                granularity = DatasetGranularity.ENTITY_LEVEL
                entity_col = None
                time_col = None

            # Extract dataset name from path
            name = self._extract_dataset_name(path)
            is_excluded = name in self._excluded_datasets

            if not include_excluded and is_excluded:
                continue

            datasets.append(
                DatasetInfo(
                    name=name,
                    findings_path=str(path),
                    source_path=findings.source_path,
                    granularity=granularity,
                    row_count=findings.row_count,
                    column_count=findings.column_count,
                    entity_column=entity_col,
                    time_column=time_col,
                    target_column=findings.target_column,
                    excluded=is_excluded,
                )
            )

        return datasets

    def create_multi_dataset_findings(
        self,
        dataset_names: Optional[List[str]] = None,
        merge_scaffold=None,
    ) -> MultiDatasetFindings:
        """Create a MultiDatasetFindings from discovered datasets.

        Args:
            dataset_names: Optional list of dataset names to include.
                          If None, all discovered datasets are included.
                          If provided, only datasets matching these names are included.
            merge_scaffold: Optional list of MergeScaffoldEntry from ProjectContext.
                           Converted to relationships for the silver layer.
        """
        datasets_info = self.list_datasets(include_excluded=True)

        # Filter to specified datasets if provided
        if dataset_names:
            datasets_info = [d for d in datasets_info if d.name in dataset_names]

        datasets = {d.name: d for d in datasets_info}
        event_datasets = [d.name for d in datasets_info if d.granularity == DatasetGranularity.EVENT_LEVEL]
        excluded = [d.name for d in datasets_info if d.excluded]

        # Determine primary entity dataset (one with target, or largest entity-level)
        primary = None
        for d in datasets_info:
            if d.granularity == DatasetGranularity.ENTITY_LEVEL:
                if d.target_column:
                    primary = d.name
                    break
                elif primary is None:
                    primary = d.name

        relationships = self._scaffold_to_relationships(merge_scaffold) if merge_scaffold else []

        return MultiDatasetFindings(
            datasets=datasets,
            relationships=relationships,
            primary_entity_dataset=primary,
            event_datasets=event_datasets,
            excluded_datasets=excluded,
        )

    @staticmethod
    def _scaffold_to_relationships(scaffold) -> List[DatasetRelationshipInfo]:
        result = []
        for entry in scaffold:
            result.append(
                DatasetRelationshipInfo(
                    left_dataset=entry.right_dataset,
                    right_dataset=entry.left_dataset,
                    left_columns=list(entry.join_keys),
                    right_columns=list(entry.join_keys),
                    relationship_type=entry.relationship,
                    confidence=1.0,
                    auto_detected=True,
                )
            )
        return result

    def exclude_dataset(self, name_pattern: str) -> None:
        """Exclude a dataset from multi-dataset analysis."""
        for dataset in self.list_datasets(include_excluded=True):
            if name_pattern.lower() in dataset.name.lower():
                self._excluded_datasets.add(dataset.name)
                return

    def include_dataset(self, name_pattern: str) -> None:
        """Re-include a previously excluded dataset."""
        to_remove = None
        for name in self._excluded_datasets:
            if name_pattern.lower() in name.lower():
                to_remove = name
                break
        if to_remove:
            self._excluded_datasets.remove(to_remove)

    def get_aggregated_path(self, original_findings_path: str) -> Optional[str]:
        """Get the aggregated findings path for an event-level dataset.

        Returns the path to the aggregated findings file if:
        - The original findings is event-level
        - Aggregation has been executed (via 01d notebook)

        Returns None if the dataset is entity-level or not yet aggregated.
        """
        findings = ExplorationFindings.load(original_findings_path)

        if not findings.has_aggregated_output:
            return None

        return findings.time_series_metadata.aggregated_findings_path

    def _extract_dataset_name(self, path: Path) -> str:
        """Extract dataset name from findings path.

        Pattern: {name}_findings.yaml or {name}_aggregated_findings.yaml
        """
        stem = path.stem.replace("_findings", "")
        if "_aggregated" in stem:
            stem = stem.rsplit("_aggregated", 1)[0]
        return stem



from pathlib import Path
from typing import Any, Optional, Tuple

import numpy as np
from sklearn.preprocessing import RobustScaler, StandardScaler

from customer_retention.analysis.auto_explorer.layered_recommendations import (
    LayeredRecommendation,
    RecommendationRegistry,
)
from customer_retention.core.compat import native_pd, safe_to_datetime


class DataMaterializer:
    def __init__(self, registry: RecommendationRegistry, output_dir: Optional[str] = None,
                 storage=None, context=None):
        self.registry = registry
        self.output_dir = output_dir or "./prepared_data"
        self.storage = storage or _get_storage()
        self.context = context

    def transform(self, df: Any) -> native_pd.DataFrame:
        df = self.apply_bronze(df)
        df = self.apply_silver(df)
        df = self.apply_gold(df)
        return df

    def materialize(self, df: native_pd.DataFrame, output_name: str) -> Tuple[native_pd.DataFrame, str]:
        result_df = self.transform(df)
        output_path = self._save(result_df, output_name)
        return result_df, output_path

    def apply_bronze(self, df: Any) -> native_pd.DataFrame:
        if not self.registry.bronze:
            return df
        df = df.copy()
        for rec in self.registry.bronze.null_handling:
            df = self._apply_null_handling(df, rec)
        for rec in self.registry.bronze.outlier_handling:
            df = self._apply_outlier_handling(df, rec)
        for rec in self.registry.bronze.type_conversions:
            df = self._apply_type_conversion(df, rec)
        for rec in self.registry.bronze.filtering:
            df = self._apply_filtering(df, rec)
        return df

    def apply_silver(self, df: Any) -> native_pd.DataFrame:
        if not self.registry.silver:
            return df
        df = df.copy()
        for rec in self.registry.silver.aggregations:
            df = self._apply_aggregation(df, rec)
        for rec in self.registry.silver.derived_columns:
            df = self._apply_derived(df, rec)
        return df

    def apply_gold(self, df: Any) -> native_pd.DataFrame:
        if not self.registry.gold:
            return df
        df = df.copy()
        for rec in self.registry.gold.transformations:
            df = self._apply_transformation(df, rec)
        for rec in self.registry.gold.encoding:
            df = self._apply_encoding(df, rec)
        for rec in self.registry.gold.scaling:
            df = self._apply_scaling(df, rec)
        return df

    def _apply_null_handling(self, df: native_pd.DataFrame, rec: LayeredRecommendation) -> native_pd.DataFrame:
        col = rec.target_column
        if col not in df.columns:
            return df
        strategy = rec.parameters.get("strategy", "median")
        if strategy == "median":
            df[col] = df[col].fillna(df[col].median())
        elif strategy == "mean":
            df[col] = df[col].fillna(df[col].mean())
        elif strategy == "mode":
            df[col] = df[col].fillna(df[col].mode().iloc[0] if len(df[col].mode()) > 0 else df[col])
        elif strategy == "zero":
            df[col] = df[col].fillna(0)
        return df

    def _apply_outlier_handling(self, df: native_pd.DataFrame, rec: LayeredRecommendation) -> native_pd.DataFrame:
        col = rec.target_column
        if col not in df.columns:
            return df
        method = rec.parameters.get("method", "iqr")
        if method == "iqr":
            factor = rec.parameters.get("factor", 1.5)
            q1, q3 = df[col].quantile([0.25, 0.75])
            iqr = q3 - q1
            lower, upper = q1 - factor * iqr, q3 + factor * iqr
            df[col] = df[col].clip(lower, upper)
        return df

    def _apply_type_conversion(self, df: native_pd.DataFrame, rec: LayeredRecommendation) -> native_pd.DataFrame:
        col = rec.target_column
        if col not in df.columns:
            return df
        target_type = rec.parameters.get("target_type", "str")
        if target_type == "datetime":
            df[col] = safe_to_datetime(df[col])
        else:
            df[col] = df[col].astype(target_type)
        return df

    def _apply_filtering(self, df: native_pd.DataFrame, rec: LayeredRecommendation) -> native_pd.DataFrame:
        col = rec.target_column
        if rec.action == "drop" and col in df.columns:
            df = df.drop(columns=[col])
        return df

    def _apply_aggregation(self, df: native_pd.DataFrame, rec: LayeredRecommendation) -> native_pd.DataFrame:
        col = rec.target_column
        if col not in df.columns or not self.registry.silver:
            return df
        entity_col = self.registry.silver.entity_column
        agg_func = rec.parameters.get("aggregation", "sum")
        feature_name = f"{col}_{agg_func}"
        df[feature_name] = df.groupby(entity_col)[col].transform(agg_func)
        return df

    def _apply_derived(self, df: native_pd.DataFrame, rec: LayeredRecommendation) -> native_pd.DataFrame:
        return df

    def _apply_transformation(self, df: native_pd.DataFrame, rec: LayeredRecommendation) -> native_pd.DataFrame:
        col = rec.target_column
        if col not in df.columns:
            return df
        method = rec.parameters.get("method", "log")
        if method == "log":
            df[col] = np.log1p(df[col])
        elif method == "sqrt":
            df[col] = np.sqrt(df[col])
        return df

    def _apply_encoding(self, df: native_pd.DataFrame, rec: LayeredRecommendation) -> native_pd.DataFrame:
        col = rec.target_column
        if col not in df.columns:
            return df
        method = rec.parameters.get("method", "one_hot")
        if method == "one_hot":
            dummies = native_pd.get_dummies(df[col], prefix=col, drop_first=False)
            df = native_pd.concat([df.drop(columns=[col]), dummies], axis=1)
        return df

    def _apply_scaling(self, df: native_pd.DataFrame, rec: LayeredRecommendation) -> native_pd.DataFrame:
        col = rec.target_column
        if col not in df.columns:
            return df
        method = rec.parameters.get("method", "standard")
        scaler = StandardScaler() if method == "standard" else RobustScaler()
        df[col] = scaler.fit_transform(df[[col]])
        return df

    def _save(self, df: native_pd.DataFrame, output_name: str) -> str:
        Path(self.output_dir).mkdir(parents=True, exist_ok=True)
        output_path = f"{self.output_dir}/{output_name}"
        if self.storage:
            metadata = self.context.build_commit_metadata() if self.context else None
            self.storage.write(df, output_path, metadata=metadata)
            if self.context:
                version = len(self.storage.history(output_path)) - 1
                self.context.delta_versions[output_path] = version
        else:
            output_path = f"{output_path}.parquet"
            df.to_parquet(output_path, index=False)
        return output_path

    def compare_runs(self, path: str, version_a: int, version_b: int) -> native_pd.DataFrame:
        if not self.storage:
            raise RuntimeError("DeltaStorage required for version comparison")
        df_a = self.storage.read(path, version=version_a)
        df_b = self.storage.read(path, version=version_b)
        return native_pd.DataFrame({
            "metric": ["row_count", "column_count"],
            "version_a": [len(df_a), len(df_a.columns)],
            "version_b": [len(df_b), len(df_b.columns)],
        })


def _get_storage():
    try:
        from customer_retention.integrations.adapters.factory import get_delta
        return get_delta()
    except ImportError:
        return None



from typing import TYPE_CHECKING, Any, Dict, List

from .layered_recommendations import LayeredRecommendation, RecommendationRegistry

if TYPE_CHECKING:
    from .findings import ExplorationFindings


class BronzeBuilder:
    def __init__(self, parent: "RecommendationBuilder"):
        self._parent = parent
        self._registry = parent.registry
        self._notebook = parent.notebook
        if self._registry.bronze is None:
            self._registry.init_bronze(parent.findings.source_path)

    def impute_nulls(self, column: str, strategy: str, reason: str) -> "BronzeBuilder":
        self._registry.add_bronze_null(column, strategy, reason, self._notebook)
        return self

    def cap_outliers(self, column: str, method: str, reason: str = "", **kwargs) -> "BronzeBuilder":
        params = {"method": method, **kwargs}
        self._registry.add_bronze_outlier(column, "cap", params, reason, self._notebook)
        return self

    def drop_column(self, column: str, reason: str) -> "BronzeBuilder":
        rec = LayeredRecommendation(
            id=f"bronze_drop_{column}", layer="bronze", category="filtering",
            action="drop", target_column=column, parameters={},
            rationale=reason, source_notebook=self._notebook
        )
        self._registry.bronze.filtering.append(rec)
        return self

    def convert_type(self, column: str, target_type: str, reason: str) -> "BronzeBuilder":
        rec = LayeredRecommendation(
            id=f"bronze_type_{column}", layer="bronze", category="type",
            action="cast", target_column=column, parameters={"target_type": target_type},
            rationale=reason, source_notebook=self._notebook
        )
        self._registry.bronze.type_conversions.append(rec)
        return self


class SilverBuilder:
    def __init__(self, parent: "RecommendationBuilder"):
        self._parent = parent
        self._registry = parent.registry
        self._notebook = parent.notebook
        if self._registry.silver is None:
            entity_col = (parent.findings.identifier_columns[0]
                          if parent.findings.identifier_columns else "id")
            time_col = (parent.findings.datetime_columns[0]
                        if parent.findings.datetime_columns else None)
            self._registry.init_silver(entity_col, time_col)

    def aggregate(self, column: str, aggregation: str, windows: List[str], reason: str) -> "SilverBuilder":
        self._registry.add_silver_aggregation(column, aggregation, windows, reason, self._notebook)
        return self

    def join(self, dataset: str, join_key: str, join_type: str, reason: str) -> "SilverBuilder":
        rec = LayeredRecommendation(
            id=f"silver_join_{dataset}", layer="silver", category="join",
            action="join", target_column=join_key,
            parameters={"dataset": dataset, "join_type": join_type},
            rationale=reason, source_notebook=self._notebook
        )
        self._registry.silver.joins.append(rec)
        return self

    def derive(self, column_name: str, formula: str, reason: str) -> "SilverBuilder":
        rec = LayeredRecommendation(
            id=f"silver_derive_{column_name}", layer="silver", category="derived",
            action="compute", target_column=column_name, parameters={"formula": formula},
            rationale=reason, source_notebook=self._notebook
        )
        self._registry.silver.derived_columns.append(rec)
        return self


class GoldBuilder:
    def __init__(self, parent: "RecommendationBuilder"):
        self._parent = parent
        self._registry = parent.registry
        self._notebook = parent.notebook
        if self._registry.gold is None:
            target = parent.findings.target_column or "target"
            self._registry.init_gold(target)

    def encode(self, column: str, method: str, reason: str, **kwargs) -> "GoldBuilder":
        rec = LayeredRecommendation(
            id=f"gold_encode_{column}", layer="gold", category="encoding",
            action=method, target_column=column, parameters={"method": method, **kwargs},
            rationale=reason, source_notebook=self._notebook
        )
        self._registry.gold.encoding.append(rec)
        return self

    def scale(self, column: str, method: str, reason: str) -> "GoldBuilder":
        rec = LayeredRecommendation(
            id=f"gold_scale_{column}", layer="gold", category="scaling",
            action=method, target_column=column, parameters={"method": method},
            rationale=reason, source_notebook=self._notebook
        )
        self._registry.gold.scaling.append(rec)
        return self

    def select(self, column: str, include: bool, reason: str) -> "GoldBuilder":
        action = "include" if include else "exclude"
        rec = LayeredRecommendation(
            id=f"gold_select_{column}", layer="gold", category="selection",
            action=action, target_column=column, parameters={"include": include},
            rationale=reason, source_notebook=self._notebook
        )
        self._registry.gold.feature_selection.append(rec)
        return self

    def transform(self, column: str, method: str, reason: str) -> "GoldBuilder":
        rec = LayeredRecommendation(
            id=f"gold_transform_{column}", layer="gold", category="transformation",
            action=method, target_column=column, parameters={"method": method},
            rationale=reason, source_notebook=self._notebook
        )
        self._registry.gold.transformations.append(rec)
        return self


class RecommendationBuilder:
    def __init__(self, findings: "ExplorationFindings", notebook: str):
        self.findings = findings
        self.notebook = notebook
        self.registry = RecommendationRegistry()

    def bronze(self) -> BronzeBuilder:
        return BronzeBuilder(self)

    def silver(self) -> SilverBuilder:
        return SilverBuilder(self)

    def gold(self) -> GoldBuilder:
        return GoldBuilder(self)

    @property
    def all_recommendations(self) -> List[LayeredRecommendation]:
        return self.registry.all_recommendations

    def to_dict(self) -> Dict[str, Any]:
        return self.registry.to_dict()



"""
Behavioral feature generation for customer retention analysis.

This module provides behavioral feature calculations such as frequency,
engagement, service adoption, and recency buckets.
"""

from dataclasses import dataclass, field
from typing import List, Optional

import numpy as np

from customer_retention.core.compat import DataFrame, cut, to_datetime


@dataclass
class BehavioralFeatureResult:
    """Result of behavioral feature generation."""
    df: DataFrame
    generated_features: List[str]
    warnings: List[str] = field(default_factory=list)
    pit_warnings: List[str] = field(default_factory=list)


class BehavioralFeatureGenerator:
    """
    Generates behavioral features from customer data.

    Behavioral features capture customer activity patterns, engagement levels,
    and service adoption metrics.

    Parameters
    ----------
    tenure_months_column : str, optional
        Column containing customer tenure in months.
    tenure_days_column : str, optional
        Column containing customer tenure in days.
    total_orders_column : str, optional
        Column containing total number of orders.
    emails_sent_column : str, optional
        Column containing number of emails sent.
    total_visits_column : str, optional
        Column containing total visits.
    days_since_last_order_column : str, optional
        Column containing days since last order.
    open_rate_column : str, optional
        Column containing email open rate.
    click_rate_column : str, optional
        Column containing email click rate.
    service_columns : List[str], optional
        List of columns indicating service adoption (binary).
    recency_bins : List[float], optional
        Bin edges for recency buckets.
    recency_labels : List[str], optional
        Labels for recency buckets.

    Attributes
    ----------
    generated_features : List[str]
        Names of features generated during last transform.
    """

    DEFAULT_RECENCY_BINS = [0, 7, 30, 60, 90, float('inf')]
    DEFAULT_RECENCY_LABELS = ["active", "recent", "warm", "cooling", "dormant"]

    def __init__(
        self,
        tenure_months_column: Optional[str] = None,
        tenure_days_column: Optional[str] = None,
        total_orders_column: Optional[str] = None,
        emails_sent_column: Optional[str] = None,
        total_visits_column: Optional[str] = None,
        days_since_last_order_column: Optional[str] = None,
        open_rate_column: Optional[str] = None,
        click_rate_column: Optional[str] = None,
        service_columns: Optional[List[str]] = None,
        recency_bins: Optional[List[float]] = None,
        recency_labels: Optional[List[str]] = None,
        enforce_point_in_time: bool = False,
        feature_timestamp_column: Optional[str] = None,
    ):
        self.tenure_months_column = tenure_months_column
        self.tenure_days_column = tenure_days_column
        self.total_orders_column = total_orders_column
        self.emails_sent_column = emails_sent_column
        self.total_visits_column = total_visits_column
        self.days_since_last_order_column = days_since_last_order_column
        self.open_rate_column = open_rate_column
        self.click_rate_column = click_rate_column
        self.service_columns = service_columns or []
        self.recency_bins = recency_bins or self.DEFAULT_RECENCY_BINS
        self.recency_labels = recency_labels or self.DEFAULT_RECENCY_LABELS
        self.enforce_point_in_time = enforce_point_in_time
        self.feature_timestamp_column = feature_timestamp_column or "feature_timestamp"

        self.generated_features: List[str] = []
        self.pit_warnings: List[str] = []
        self._is_fitted = False

    def fit(self, df: DataFrame) -> "BehavioralFeatureGenerator":
        """
        Fit the generator (stores configuration but no learning required).

        Parameters
        ----------
        df : DataFrame
            Input DataFrame.

        Returns
        -------
        self
        """
        self._is_fitted = True
        return self

    def transform(self, df: DataFrame) -> DataFrame:
        """
        Generate behavioral features for the input DataFrame.

        Parameters
        ----------
        df : DataFrame
            Input DataFrame.

        Returns
        -------
        DataFrame
            DataFrame with behavioral features added.
        """
        if not self._is_fitted:
            raise ValueError("Generator not fitted. Call fit() first.")

        result = df.copy()
        self.generated_features = []
        self.pit_warnings = []

        if self.enforce_point_in_time:
            self._validate_point_in_time(result)

        # Frequency features
        result = self._generate_frequency_features(result)

        # Engagement features
        result = self._generate_engagement_features(result)

        # Service adoption features
        result = self._generate_service_adoption_features(result)

        # Recency bucket
        result = self._generate_recency_bucket(result)

        return result

    def fit_transform(self, df: DataFrame) -> DataFrame:
        """
        Fit and transform in one step.

        Parameters
        ----------
        df : DataFrame
            Input DataFrame.

        Returns
        -------
        DataFrame
            DataFrame with behavioral features added.
        """
        self.fit(df)
        return self.transform(df)

    def _generate_frequency_features(self, df: DataFrame) -> DataFrame:
        """Generate frequency-based features."""
        # Order frequency
        if self.tenure_months_column and self.total_orders_column:
            if self.tenure_months_column in df.columns and self.total_orders_column in df.columns:
                tenure = df[self.tenure_months_column].replace(0, np.nan)
                df["order_frequency"] = df[self.total_orders_column] / tenure
                self.generated_features.append("order_frequency")

        # Email frequency
        if self.tenure_months_column and self.emails_sent_column:
            if self.tenure_months_column in df.columns and self.emails_sent_column in df.columns:
                tenure = df[self.tenure_months_column].replace(0, np.nan)
                df["email_frequency"] = df[self.emails_sent_column] / tenure
                self.generated_features.append("email_frequency")

        # Visit frequency
        if self.tenure_months_column and self.total_visits_column:
            if self.tenure_months_column in df.columns and self.total_visits_column in df.columns:
                tenure = df[self.tenure_months_column].replace(0, np.nan)
                df["visit_frequency"] = df[self.total_visits_column] / tenure
                self.generated_features.append("visit_frequency")

        # Order recency ratio
        if self.tenure_days_column and self.days_since_last_order_column:
            if self.tenure_days_column in df.columns and self.days_since_last_order_column in df.columns:
                tenure = df[self.tenure_days_column].replace(0, np.nan)
                df["order_recency_ratio"] = df[self.days_since_last_order_column] / tenure
                self.generated_features.append("order_recency_ratio")

        return df

    def _generate_engagement_features(self, df: DataFrame) -> DataFrame:
        """Generate engagement-based features."""
        if self.open_rate_column and self.click_rate_column:
            if self.open_rate_column in df.columns and self.click_rate_column in df.columns:
                # Email engagement score
                df["email_engagement_score"] = (
                    df[self.open_rate_column] + df[self.click_rate_column]
                ) / 2
                self.generated_features.append("email_engagement_score")

                # Click to open rate (handle division by zero)
                open_rate = df[self.open_rate_column].replace(0, np.nan)
                df["click_to_open_rate"] = df[self.click_rate_column] / open_rate
                df["click_to_open_rate"] = df["click_to_open_rate"].fillna(0)
                self.generated_features.append("click_to_open_rate")

        return df

    def _generate_service_adoption_features(self, df: DataFrame) -> DataFrame:
        """Generate service adoption features."""
        if self.service_columns:
            # Check which columns exist
            existing_cols = [c for c in self.service_columns if c in df.columns]
            if existing_cols:
                # Service adoption score (count of services)
                df["service_adoption_score"] = df[existing_cols].sum(axis=1).astype(float)
                self.generated_features.append("service_adoption_score")

                # Service adoption percentage
                df["service_adoption_pct"] = df[existing_cols].sum(axis=1) / len(existing_cols)
                self.generated_features.append("service_adoption_pct")

        return df

    def _generate_recency_bucket(self, df: DataFrame) -> DataFrame:
        """Generate recency bucket feature."""
        if self.days_since_last_order_column:
            if self.days_since_last_order_column in df.columns:
                df["recency_bucket"] = cut(
                    df[self.days_since_last_order_column],
                    bins=self.recency_bins,
                    labels=self.recency_labels,
                    include_lowest=True
                )
                self.generated_features.append("recency_bucket")

        return df

    def _validate_point_in_time(self, df: DataFrame) -> None:
        """Validate that behavioral inputs respect point-in-time constraints."""
        if self.feature_timestamp_column not in df.columns:
            return

        feature_ts = to_datetime(df[self.feature_timestamp_column], format='mixed')
        datetime_cols = df.select_dtypes(include=["datetime64"]).columns

        for col in datetime_cols:
            if col == self.feature_timestamp_column:
                continue
            violations = df[df[col] > feature_ts]
            if len(violations) > 0:
                self.pit_warnings.append(
                    f"PIT Warning: {len(violations)} rows have {col} > {self.feature_timestamp_column}"
                )



"""Cross-validation strategies for model evaluation."""

import gc
import time
from dataclasses import dataclass
from enum import Enum
from typing import Any, Callable, Dict, List, Optional, Tuple

import numpy as np
from sklearn.model_selection import GroupKFold, RepeatedStratifiedKFold, StratifiedKFold, cross_val_score

from customer_retention.core.compat import DataFrame, Series, collect_for_sklearn, native_pd

_CV_ENTITY_COL = "__cv_entity__"
_CV_DATE_COL = "__cv_date__"


def _squeeze_to_series(obj: Any) -> Any:
    if isinstance(obj, native_pd.DataFrame) and obj.shape[1] == 1:
        return obj.iloc[:, 0]
    return obj


def _has_cv_metadata(X: Any) -> bool:
    return _CV_ENTITY_COL in X.columns and _CV_DATE_COL in X.columns


class CVStrategy(Enum):
    STRATIFIED_KFOLD = "stratified_kfold"
    REPEATED_STRATIFIED = "repeated_stratified"
    TIME_SERIES = "time_series"
    GROUP_KFOLD = "group_kfold"
    TEMPORAL_ENTITY = "temporal_entity"


class TemporalEntitySplit:
    def __init__(self, n_splits: int = 5, temporal_values: Optional[Series] = None, purge_gap_days: int = 0):
        self.n_splits = n_splits
        self.temporal_values = temporal_values
        self.purge_gap_days = purge_gap_days
        self._inner = GroupKFold(n_splits=n_splits)

    def get_n_splits(self, X=None, y=None, groups=None):
        return self.n_splits

    def split(self, X, y=None, groups=None):
        for train_idx, test_idx in self._inner.split(X, y, groups):
            if self.purge_gap_days > 0 and self.temporal_values is not None:
                test_dates = self.temporal_values.iloc[test_idx]
                test_min = native_pd.Timestamp(test_dates.min())
                purge_cutoff = test_min - native_pd.Timedelta(days=self.purge_gap_days)
                train_dates = self.temporal_values.iloc[train_idx]
                keep_mask = train_dates < purge_cutoff
                purged_train = train_idx[keep_mask.to_numpy()]
                if len(purged_train) > len(train_idx) * 0.1:
                    train_idx = purged_train
            yield train_idx, test_idx


@dataclass
class CVResult:
    cv_scores: np.ndarray
    cv_mean: float
    cv_std: float
    fold_details: List[Dict[str, Any]]
    scoring: str
    is_stable: bool


class CrossValidator:
    def __init__(
        self,
        strategy: CVStrategy = CVStrategy.STRATIFIED_KFOLD,
        n_splits: int = 5,
        n_repeats: int = 1,
        shuffle: bool = True,
        random_state: int = 42,
        scoring: str = "average_precision",
        stability_threshold: float = 0.10,
        purge_gap_days: int = 0,
    ):
        self.strategy = strategy
        self.n_splits = n_splits
        self.n_repeats = n_repeats
        self.shuffle = shuffle
        self.random_state = random_state
        self.scoring = scoring
        self.stability_threshold = stability_threshold
        self.purge_gap_days = purge_gap_days

    @staticmethod
    def _extract_importance(model, feature_names: List[str]) -> Optional[Dict[str, float]]:
        if hasattr(model, 'feature_importances_'):
            return dict(zip(feature_names, model.feature_importances_.tolist()))
        if hasattr(model, 'coef_'):
            return dict(zip(feature_names, np.abs(model.coef_).flatten().tolist()))
        return None

    def run(
        self,
        model,
        X: DataFrame,
        y: Series,
        groups: Optional[Series] = None,
        temporal_values: Optional[Series] = None,
        on_fold_complete: Optional[Callable] = None,
        capture_fold_importance: bool = False,
        feature_names: Optional[List[str]] = None,
    ) -> CVResult:
        _capture = capture_fold_importance and feature_names is not None
        if hasattr(model, "clone") and self.strategy == CVStrategy.TEMPORAL_ENTITY:
            return self._run_distributed(
                model, X, y, groups=groups, temporal_values=temporal_values,
                on_fold_complete=on_fold_complete,
                capture_fold_importance=_capture, feature_names=feature_names,
            )

        X, y = collect_for_sklearn(X), collect_for_sklearn(y)
        if groups is not None:
            groups = _squeeze_to_series(collect_for_sklearn(groups))
        if temporal_values is not None:
            temporal_values = _squeeze_to_series(collect_for_sklearn(temporal_values))
        cv_splitter = self._create_cv_splitter(groups, temporal_values)
        fold_details = []

        if self.strategy == CVStrategy.TEMPORAL_ENTITY or on_fold_complete is not None or _capture:
            scores, fold_details = self._run_manual_cv(
                model, X, y, cv_splitter, groups=groups, on_fold_complete=on_fold_complete,
                capture_fold_importance=_capture, feature_names=feature_names,
            )
        elif self.strategy == CVStrategy.GROUP_KFOLD:
            scores = cross_val_score(model, X, y, cv=cv_splitter, scoring=self.scoring, groups=groups)
            fold_details = self._collect_fold_details_with_groups(X, y, groups, cv_splitter)
        else:
            scores = cross_val_score(model, X, y, cv=cv_splitter, scoring=self.scoring)
            fold_details = self._collect_fold_details(X, y, cv_splitter)

        cv_mean = np.mean(scores)
        cv_std = np.std(scores)
        is_stable = bool(cv_std <= self.stability_threshold)

        return CVResult(
            cv_scores=scores,
            cv_mean=cv_mean,
            cv_std=cv_std,
            fold_details=fold_details,
            scoring=self.scoring,
            is_stable=is_stable,
        )

    def _run_distributed(
        self,
        model,
        X: DataFrame,
        y: Series,
        groups: Optional[Series] = None,
        temporal_values: Optional[Series] = None,
        on_fold_complete: Optional[Callable] = None,
        capture_fold_importance: bool = False,
        feature_names: Optional[List[str]] = None,
    ) -> CVResult:
        from customer_retention.core.compat import _is_spark_pandas

        groups_pd = _squeeze_to_series(collect_for_sklearn(groups)) if groups is not None else None
        temporal_pd = _squeeze_to_series(collect_for_sklearn(temporal_values)) if temporal_values is not None else None

        n_rows = len(groups_pd) if groups_pd is not None else len(y)
        splitter = TemporalEntitySplit(
            n_splits=self.n_splits,
            temporal_values=temporal_pd,
            purge_gap_days=self.purge_gap_days,
        )
        all_folds = list(splitter.split(np.zeros(n_rows), groups=groups_pd))

        if _is_spark_pandas(X) and _has_cv_metadata(X):
            return self._score_folds_spark(
                model, X, y, groups_pd, temporal_pd, all_folds,
                on_fold_complete=on_fold_complete,
                capture_fold_importance=capture_fold_importance,
                feature_names=feature_names,
            )

        X_pd = collect_for_sklearn(X)
        y_pd = _squeeze_to_series(collect_for_sklearn(y))
        return self._score_folds_pandas(
            model, X_pd, y_pd, groups_pd, all_folds,
            on_fold_complete=on_fold_complete,
            capture_fold_importance=capture_fold_importance,
            feature_names=feature_names,
        )

    def _score_fold(self, y_true: np.ndarray, y_proba: np.ndarray) -> float:
        from sklearn.metrics import average_precision_score, f1_score, roc_auc_score

        if self.scoring == "roc_auc":
            return roc_auc_score(y_true, y_proba[:, 1])
        if self.scoring in ("average_precision", "pr_auc"):
            return average_precision_score(y_true, y_proba[:, 1])
        if self.scoring == "f1_weighted":
            y_pred = np.argmax(y_proba, axis=1)
            return f1_score(y_true, y_pred, average="weighted")
        return roc_auc_score(y_true, y_proba[:, 1])

    def _run_manual_cv(
        self,
        model,
        X,
        y,
        cv_splitter,
        groups: Optional[Series] = None,
        on_fold_complete: Optional[Callable] = None,
        capture_fold_importance: bool = False,
        feature_names: Optional[List[str]] = None,
    ) -> Tuple[np.ndarray, List[Dict[str, Any]]]:
        from sklearn.base import clone

        scores = []
        fold_details = []
        total_folds = cv_splitter.get_n_splits(X, y, groups)

        for fold_idx, (train_idx, test_idx) in enumerate(cv_splitter.split(X, y, groups)):
            t0 = time.monotonic()
            fold_model = clone(model)
            fold_model.fit(X.iloc[train_idx], y.iloc[train_idx])
            y_proba = fold_model.predict_proba(X.iloc[test_idx])
            y_fold_test = y.iloc[test_idx].to_numpy()
            score = self._score_fold(y_fold_test, y_proba)
            elapsed = time.monotonic() - t0

            scores.append(score)
            detail: Dict[str, Any] = {
                "fold": fold_idx + 1,
                "train_size": len(train_idx),
                "test_size": len(test_idx),
                "train_class_ratio": float(y.iloc[train_idx].mean()),
                "score": score,
                "elapsed_seconds": elapsed,
            }
            if groups is not None:
                detail["train_entities"] = int(groups.iloc[train_idx].nunique())
                detail["test_entities"] = int(groups.iloc[test_idx].nunique())
            if capture_fold_importance and feature_names:
                imp = self._extract_importance(fold_model, feature_names)
                if imp is not None:
                    detail["feature_importance"] = imp
            fold_details.append(detail)
            del fold_model, y_proba, y_fold_test
            gc.collect()

            if on_fold_complete is not None:
                on_fold_complete(detail, fold_idx + 1, total_folds)

        return np.array(scores), fold_details

    def _score_folds_pandas(self, model, X_pd, y_pd, groups_pd, all_folds,
                            on_fold_complete=None, capture_fold_importance=False,
                            feature_names=None) -> CVResult:
        scores = []
        fold_details = []
        total_folds = len(all_folds)

        for fold_idx, (train_idx, test_idx) in enumerate(all_folds):
            t0 = time.monotonic()
            fold_model = model.clone()
            fold_model.fit(X_pd.iloc[train_idx], y_pd.iloc[train_idx])
            y_proba = fold_model.predict_proba(X_pd.iloc[test_idx])
            y_fold_test = y_pd.iloc[test_idx].to_numpy()

            score = self._score_fold(y_fold_test, y_proba)
            elapsed = time.monotonic() - t0
            scores.append(score)

            detail: Dict[str, Any] = {
                "fold": fold_idx + 1,
                "train_size": len(train_idx),
                "test_size": len(test_idx),
                "train_class_ratio": float(y_pd.iloc[train_idx].mean()),
                "score": score,
                "elapsed_seconds": elapsed,
            }
            if groups_pd is not None:
                detail["train_entities"] = int(groups_pd.iloc[train_idx].nunique())
                detail["test_entities"] = int(groups_pd.iloc[test_idx].nunique())
            if capture_fold_importance and feature_names:
                imp = self._extract_importance(fold_model, feature_names)
                if imp is not None:
                    detail["feature_importance"] = imp
            fold_details.append(detail)
            del fold_model, y_proba, y_fold_test
            gc.collect()

            if on_fold_complete is not None:
                on_fold_complete(detail, fold_idx + 1, total_folds)

        return self._build_cv_result(scores, fold_details)

    def _score_folds_spark(self, model, X, y, groups_pd, temporal_pd, all_folds,
                           on_fold_complete=None, capture_fold_importance=False,
                           feature_names=None) -> CVResult:
        from customer_retention.core.compat import concat as compat_concat

        combined = compat_concat([X, y.rename("__y__")], axis=1)
        feature_cols = [c for c in X.columns if not c.startswith("__cv_")]

        scores = []
        fold_details = []
        total_folds = len(all_folds)

        for fold_idx, (train_idx, test_idx) in enumerate(all_folds):
            t0 = time.monotonic()
            fold_model = model.clone()

            train_entities = groups_pd.iloc[train_idx].unique().tolist()
            test_entities = groups_pd.iloc[test_idx].unique().tolist()

            train_fold = combined[combined[_CV_ENTITY_COL].isin(train_entities)]
            test_fold = combined[combined[_CV_ENTITY_COL].isin(test_entities)]

            if self.purge_gap_days > 0 and temporal_pd is not None:
                test_dates = temporal_pd.iloc[test_idx]
                test_min = native_pd.Timestamp(test_dates.min())
                purge_cutoff = test_min - native_pd.Timedelta(days=self.purge_gap_days)
                train_purged = train_fold[train_fold[_CV_DATE_COL] < purge_cutoff]
                if len(train_purged) > len(train_fold) * 0.1:
                    train_fold = train_purged

            fold_model.fit(train_fold[feature_cols], train_fold["__y__"])
            y_proba = fold_model.predict_proba(test_fold[feature_cols])
            y_fold_test = collect_for_sklearn(test_fold["__y__"]).to_numpy()

            score = self._score_fold(y_fold_test, y_proba)
            elapsed = time.monotonic() - t0
            scores.append(score)

            detail: Dict[str, Any] = {
                "fold": fold_idx + 1,
                "train_size": int(len(train_fold)),
                "test_size": int(len(test_fold)),
                "train_class_ratio": float(train_fold["__y__"].mean()),
                "score": score,
                "elapsed_seconds": elapsed,
            }
            if groups_pd is not None:
                detail["train_entities"] = int(groups_pd.iloc[train_idx].nunique())
                detail["test_entities"] = int(groups_pd.iloc[test_idx].nunique())
            if capture_fold_importance:
                _fn = feature_names or feature_cols
                imp = self._extract_importance(fold_model, _fn)
                if imp is not None:
                    detail["feature_importance"] = imp
            fold_details.append(detail)

            del fold_model, train_fold, test_fold, y_proba, y_fold_test
            gc.collect()

            if on_fold_complete is not None:
                on_fold_complete(detail, fold_idx + 1, total_folds)

        return self._build_cv_result(scores, fold_details)

    def _build_cv_result(self, scores: list, fold_details: list) -> CVResult:
        cv_scores = np.array(scores)
        cv_mean = float(np.mean(cv_scores))
        cv_std = float(np.std(cv_scores))
        return CVResult(
            cv_scores=cv_scores,
            cv_mean=cv_mean,
            cv_std=cv_std,
            fold_details=fold_details,
            scoring=self.scoring,
            is_stable=bool(cv_std <= self.stability_threshold),
        )

    def _create_cv_splitter(self, groups: Optional[Series] = None, temporal_values: Optional[Series] = None):
        if self.strategy == CVStrategy.STRATIFIED_KFOLD:
            return StratifiedKFold(n_splits=self.n_splits, shuffle=self.shuffle, random_state=self.random_state)

        if self.strategy == CVStrategy.REPEATED_STRATIFIED:
            return RepeatedStratifiedKFold(n_splits=self.n_splits, n_repeats=self.n_repeats, random_state=self.random_state)

        if self.strategy == CVStrategy.GROUP_KFOLD:
            return GroupKFold(n_splits=self.n_splits)

        if self.strategy == CVStrategy.TEMPORAL_ENTITY:
            return TemporalEntitySplit(
                n_splits=self.n_splits,
                temporal_values=temporal_values,
                purge_gap_days=self.purge_gap_days,
            )

        if self.strategy == CVStrategy.TIME_SERIES:
            from sklearn.model_selection import TimeSeriesSplit
            return TimeSeriesSplit(n_splits=self.n_splits)

        return StratifiedKFold(n_splits=self.n_splits, shuffle=self.shuffle, random_state=self.random_state)

    def _collect_fold_details(self, X: DataFrame, y: Series, cv_splitter) -> List[Dict[str, Any]]:
        fold_details = []
        for fold_idx, (train_idx, test_idx) in enumerate(cv_splitter.split(X, y)):
            y_train = y.iloc[train_idx]
            fold_details.append({
                "fold": fold_idx + 1,
                "train_size": len(train_idx),
                "test_size": len(test_idx),
                "train_class_ratio": y_train.mean(),
                "score": None,
            })
        return fold_details

    def _collect_fold_details_with_groups(
        self,
        X: DataFrame,
        y: Series,
        groups: Series,
        cv_splitter,
    ) -> List[Dict[str, Any]]:
        fold_details = []
        for fold_idx, (train_idx, test_idx) in enumerate(cv_splitter.split(X, y, groups)):
            y_train = y.iloc[train_idx]
            fold_details.append({
                "fold": fold_idx + 1,
                "train_size": len(train_idx),
                "test_size": len(test_idx),
                "train_class_ratio": y_train.mean(),
                "score": None,
            })
        return fold_details



#!/usr/bin/env python3
"""Capture non-chart cell outputs from executed Jupyter notebooks into a structured Markdown manifest.

The manifest is designed for diffing between two exploration runs to produce
a drift report.  Charts (Plotly, matplotlib images) are excluded; text,
HTML tables, and stream output are preserved with cell-ID and @cr: tag
annotations so the comparison tool can match cells across versions.

Usage:
    python scripts/experiments/capture_notebook_outputs.py \
        --notebooks-dir exploration_notebooks \
        --output experiments/runs/my-run/cell_outputs.md \
        [--notebooks 00,01,08]
"""

from __future__ import annotations

import argparse
import json
import re
import sys
from pathlib import Path
from typing import Optional

# ---------------------------------------------------------------------------
# Constants
# ---------------------------------------------------------------------------

NOTEBOOKS_ORDER = [
    "00_start_here",
    "01_data_discovery",
    "01a_temporal_deep_dive",
    "01a_a_temporal_text_deep_dive",
    "01b_temporal_quality",
    "01c_temporal_patterns",
    "01d_event_aggregation",
    "02_source_integrity",
    "03_dataset_merge",
    "04_column_deep_dive",
    "04a_text_columns_deep_dive",
    "05_relationship_analysis",
    "06_feature_opportunities",
    "07_modeling_readiness",
    "08_baseline_experiments",
    "09_business_alignment",
    "10_spec_generation",
    "11_scoring_validation",
    "12_view_documentation",
]

# MIME types that indicate a chart / image (excluded from capture)
_CHART_MIMES = frozenset(
    {
        "image/png",
        "image/jpeg",
        "image/svg+xml",
        "application/vnd.plotly.v1+json",
    }
)

_MAX_OUTPUT_LINES = 200

# Regex for @cr: tags in code cells
_CR_CODE_RE = re.compile(
    r"^#\s*@cr:(?P<kind>code|config)\s+name='(?P<name>[^']+)'\s+id=(?P<id>\w+)"
)
# Regex for cr:doc tags in markdown cells
_CR_DOC_RE = re.compile(
    r"\[//\]:\s*#\s*\(cr:doc\s+name='(?P<name>[^']+)'\s+id=(?P<id>\w+)\)"
)


# ---------------------------------------------------------------------------
# Helpers
# ---------------------------------------------------------------------------


def _parse_cr_tag(cell: dict) -> Optional[dict]:
    """Extract @cr: / cr:doc tag from the first source line of a cell."""
    source = cell.get("source", [])
    if not source:
        return None
    first_line = source[0] if isinstance(source, list) else source.split("\n")[0]
    m = _CR_CODE_RE.match(first_line)
    if m:
        return {"kind": m.group("kind"), "name": m.group("name"), "id": m.group("id")}
    m = _CR_DOC_RE.search(first_line)
    if m:
        return {"kind": "doc", "name": m.group("name"), "id": m.group("id")}
    return None


def _cell_id(cell: dict) -> str:
    """Return the stable cell ID, falling back to 'unknown'."""
    return cell.get("id", "unknown")


def _truncate(text: str, max_lines: int = _MAX_OUTPUT_LINES) -> str:
    """Truncate text to *max_lines* and append a marker if needed."""
    lines = text.split("\n")
    if len(lines) <= max_lines:
        return text
    kept = "\n".join(lines[:max_lines])
    return f"{kept}\n[...truncated {len(lines) - max_lines} lines]"


def _is_chart_only(output: dict) -> bool:
    """Return True if the output contains only chart MIME types."""
    if output.get("output_type") == "stream":
        return False
    if output.get("output_type") == "error":
        return False
    data = output.get("data", {})
    if not data:
        return False
    return all(mime in _CHART_MIMES for mime in data)


def _join_text(fragments: list | str) -> str:
    if isinstance(fragments, list):
        return "".join(fragments)
    return fragments


# ---------------------------------------------------------------------------
# Core capture logic
# ---------------------------------------------------------------------------


def capture_notebook(nb_path: Path) -> Optional[str]:
    """Capture cell outputs from a single executed notebook.

    Returns a Markdown string, or None if the notebook has no outputs.
    """
    with open(nb_path, encoding="utf-8") as f:
        nb = json.load(f)

    cells = nb.get("cells", [])
    parts: list[str] = []
    current_section: Optional[str] = None
    has_any_output = False

    for cell in cells:
        tag = _parse_cr_tag(cell)
        cid = _cell_id(cell)

        # Track the current section from markdown doc tags
        if cell["cell_type"] == "markdown" and tag and tag["kind"] == "doc":
            current_section = tag["name"]
            continue

        # Only process code cells
        if cell["cell_type"] != "code":
            continue

        outputs = cell.get("outputs", [])
        if not outputs:
            continue

        # Build cell header
        cell_name = tag["name"] if tag else f"cell_{cid}"
        cell_kind = tag["kind"] if tag else "code"
        section_ctx = f" (section: {current_section})" if current_section else ""
        header = f"## {cell_kind}: {cell_name} (id: {cid}){section_ctx}"

        cell_parts: list[str] = []

        for output in outputs:
            otype = output.get("output_type", "")

            if _is_chart_only(output):
                continue

            if otype == "stream":
                stream_name = output.get("name", "stdout")
                text = _join_text(output.get("text", ""))
                text = _truncate(text.rstrip())
                if text:
                    cell_parts.append(f"### Stream ({stream_name})\n```\n{text}\n```")

            elif otype == "error":
                ename = output.get("ename", "Error")
                evalue = output.get("evalue", "")
                tb = _join_text(output.get("traceback", []))
                # Strip ANSI escape codes from traceback
                tb = re.sub(r"\x1b\[[0-9;]*m", "", tb)
                tb = _truncate(tb.rstrip())
                cell_parts.append(
                    f"### Error: {ename}\n```\n{evalue}\n\n{tb}\n```"
                )

            elif otype in ("execute_result", "display_data"):
                data = output.get("data", {})

                if "text/html" in data:
                    html = _join_text(data["text/html"])
                    html = _truncate(html.rstrip())
                    cell_parts.append(f"### Output (text/html)\n{html}")
                elif "text/markdown" in data:
                    md = _join_text(data["text/markdown"])
                    md = _truncate(md.rstrip())
                    if md:
                        cell_parts.append(f"### Output (text/markdown)\n{md}")
                    elif "text/plain" in data:
                        text = _join_text(data["text/plain"])
                        text = _truncate(text.rstrip())
                        if text:
                            cell_parts.append(f"### Output (text/plain)\n```\n{text}\n```")
                elif "text/plain" in data:
                    text = _join_text(data["text/plain"])
                    text = _truncate(text.rstrip())
                    if text:
                        cell_parts.append(f"### Output (text/plain)\n```\n{text}\n```")

        if cell_parts:
            has_any_output = True
            parts.append(header)
            parts.extend(cell_parts)

    if not has_any_output:
        return None
    return "\n\n".join(parts)


def capture_all_notebooks(
    notebooks_dir: Path,
    filter_stems: Optional[set[str]] = None,
) -> str:
    """Capture outputs from all exploration notebooks in order.

    Returns the full Markdown manifest including a section map header.
    """
    manifest_parts: list[str] = []
    section_map_lines: list[str] = ["<!-- Section Map -->"]
    notebooks_found: list[tuple[str, Path]] = []

    for stem in NOTEBOOKS_ORDER:
        if filter_stems and stem not in filter_stems:
            continue
        # Notebooks may have a leading `-1_` prefix in some cases
        candidates = list(notebooks_dir.glob(f"{stem}.ipynb")) + list(
            notebooks_dir.glob(f"-1_{stem}.ipynb")
        )
        if not candidates:
            continue
        nb_path = candidates[0]
        notebooks_found.append((stem, nb_path))

    # Also pick up any notebooks not in the standard order
    known_stems = {s for s, _ in notebooks_found}
    for nb_path in sorted(notebooks_dir.glob("*.ipynb")):
        stem = nb_path.stem
        if stem.startswith("-1_"):
            stem = stem[3:]
        if stem not in known_stems and (not filter_stems or stem in filter_stems):
            notebooks_found.append((stem, nb_path))

    for stem, nb_path in notebooks_found:
        try:
            nb = json.load(open(nb_path, encoding="utf-8"))
        except (json.JSONDecodeError, OSError):
            section_map_lines.append(f"<!-- {stem}: UNREADABLE -->")
            continue
        cell_count = len(nb.get("cells", []))
        output_count = sum(
            1
            for c in nb.get("cells", [])
            if c.get("outputs")
        )
        section_map_lines.append(
            f"<!-- {stem} ({cell_count} cells, {output_count} with output) -->"
        )

        captured = capture_notebook(nb_path)
        if captured:
            manifest_parts.append(f"# {stem}\n\n{captured}")
        else:
            manifest_parts.append(f"# {stem}\n\n_No captured outputs._")

    header = "\n".join(section_map_lines)
    body = "\n\n---\n\n".join(manifest_parts)
    return f"{header}\n\n{body}\n"


# ---------------------------------------------------------------------------
# CLI
# ---------------------------------------------------------------------------


def main() -> None:
    parser = argparse.ArgumentParser(
        description="Capture non-chart cell outputs from executed notebooks."
    )
    parser.add_argument(
        "--notebooks-dir",
        type=Path,
        required=True,
        help="Directory containing executed .ipynb files",
    )
    parser.add_argument(
        "--output",
        type=Path,
        default=None,
        help="Output Markdown file (default: stdout)",
    )
    parser.add_argument(
        "--notebooks",
        type=str,
        default=None,
        help="Comma-separated notebook stems to capture (e.g. 00,01,08)",
    )
    args = parser.parse_args()

    notebooks_dir = args.notebooks_dir.resolve()
    if not notebooks_dir.is_dir():
        print(f"Error: {notebooks_dir} is not a directory", file=sys.stderr)
        sys.exit(1)

    filter_stems: Optional[set[str]] = None
    if args.notebooks:
        raw = [s.strip() for s in args.notebooks.split(",")]
        # Expand short names like "08" to match "08_baseline_experiments"
        filter_stems = set()
        for r in raw:
            matched = [s for s in NOTEBOOKS_ORDER if s.startswith(r)]
            if matched:
                filter_stems.update(matched)
            else:
                filter_stems.add(r)

    manifest = capture_all_notebooks(notebooks_dir, filter_stems)

    if args.output:
        args.output.parent.mkdir(parents=True, exist_ok=True)
        args.output.write_text(manifest, encoding="utf-8")
        print(f"Manifest written to {args.output}", file=sys.stderr)
    else:
        print(manifest)


if __name__ == "__main__":
    main()



import time
from dataclasses import dataclass, field
from enum import Enum
from typing import TYPE_CHECKING, Any, Callable, Dict, List, Optional

import numpy as np

from customer_retention.core.compat import (
    DataFrame,
    _numeric_column_names,
    batched_corr_matrix,
    bulk_variance,
    isna,
)

if TYPE_CHECKING:
    from customer_retention.analysis.auto_explorer.findings import FeatureAvailabilityMetadata


class SelectionMethod(Enum):
    VARIANCE = "VARIANCE"
    CORRELATION = "CORRELATION"
    MUTUAL_INFO = "MUTUAL_INFO"
    IMPORTANCE = "IMPORTANCE"
    RECURSIVE = "RECURSIVE"
    L1_SELECTION = "L1_SELECTION"


@dataclass
class FeatureSelectionResult:
    df: DataFrame
    selected_features: List[str]
    dropped_features: List[str]
    drop_reasons: Dict[str, str]
    method_used: SelectionMethod
    importance_scores: Optional[Dict[str, float]] = None


@dataclass
class AvailabilityRecommendation:
    column: str
    issue_type: str
    coverage_pct: float
    first_valid_date: Optional[str]
    last_valid_date: Optional[str]
    options: List[Dict[str, Any]] = field(default_factory=list)

    def to_dict(self) -> Dict[str, Any]:
        return {
            "column": self.column,
            "issue_type": self.issue_type,
            "coverage_pct": self.coverage_pct,
            "first_valid_date": self.first_valid_date,
            "last_valid_date": self.last_valid_date,
            "options": self.options,
        }


class FeatureSelector:
    def __init__(self, method: SelectionMethod = SelectionMethod.VARIANCE, variance_threshold: float = 0.01, correlation_threshold: float = 0.95, target_column: Optional[str] = None, preserve_features: Optional[List[str]] = None, max_features: Optional[int] = None, apply_correlation_filter: bool = False, precomputed_corr_matrix: Optional[Any] = None, l1_C: float = 1.0, l1_ratio: float = 1.0, progress_fn: Optional[Callable[[str], None]] = None, precomputed_variances: Optional[Any] = None, precomputed_medians: Optional[Dict[str, float]] = None, precomputed_non_null: Optional[Dict[str, int]] = None):
        self.method = method
        self.variance_threshold = variance_threshold
        self.correlation_threshold = correlation_threshold
        self.target_column = target_column
        self.preserve_features = preserve_features or []
        self.max_features = max_features
        self.apply_correlation_filter = apply_correlation_filter
        self._precomputed_corr_matrix = precomputed_corr_matrix
        self.l1_C = l1_C
        self.l1_ratio = l1_ratio
        self._progress_fn = progress_fn
        self._precomputed_medians = precomputed_medians
        self._precomputed_non_null = precomputed_non_null

        self.selected_features: List[str] = []
        self.dropped_features: List[str] = []
        self.drop_reasons: Dict[str, str] = {}
        self.importance_scores: Optional[Dict[str, float]] = None
        self._is_fitted = False
        self._cached_variances: Optional[Any] = precomputed_variances

    def fit(self, df: DataFrame) -> "FeatureSelector":
        feature_cols = [c for c in df.columns if c != self.target_column]
        numeric_set = _numeric_column_names(df, feature_cols)

        self.selected_features = feature_cols.copy()
        self.dropped_features = []
        self.drop_reasons = {}

        if self.method == SelectionMethod.VARIANCE:
            self._apply_variance_selection(df, feature_cols, numeric_set)
        elif self.method == SelectionMethod.CORRELATION:
            self._apply_correlation_selection(df, feature_cols, numeric_set)
        elif self.method == SelectionMethod.L1_SELECTION:
            self._apply_l1_selection(df, feature_cols, numeric_set)

        if self.apply_correlation_filter and self.method != SelectionMethod.CORRELATION:
            self._apply_correlation_selection(df, self.selected_features.copy(), numeric_set)

        if self.max_features and len(self.selected_features) > self.max_features:
            variances = self._get_variances(df, self.selected_features, numeric_set)
            to_keep = set(variances.sort_values(ascending=False).head(self.max_features).index.tolist())
            for feature in [f for f in self.selected_features if f not in to_keep and f not in self.preserve_features]:
                self.selected_features.remove(feature)
                self.dropped_features.append(feature)
                self.drop_reasons[feature] = "max_features limit"

        self._is_fitted = True
        return self

    def transform(self, df: DataFrame) -> FeatureSelectionResult:
        if not self._is_fitted:
            raise ValueError("Selector not fitted. Call fit() first.")

        cols_to_keep = self.selected_features.copy()
        if self.target_column and self.target_column in df.columns:
            cols_to_keep.append(self.target_column)

        cols_to_keep = [c for c in cols_to_keep if c in df.columns]
        result_df = df[cols_to_keep].copy()

        return FeatureSelectionResult(
            df=result_df,
            selected_features=self.selected_features.copy(),
            dropped_features=self.dropped_features.copy(),
            drop_reasons=self.drop_reasons.copy(),
            method_used=self.method,
            importance_scores=dict(self.importance_scores) if self.importance_scores else None,
        )

    def fit_transform(self, df: DataFrame) -> FeatureSelectionResult:
        self.fit(df)
        return self.transform(df)

    def _get_variances(self, df: DataFrame, features: List[str], numeric_set: set) -> Any:
        numeric_features = [f for f in features if f in numeric_set]
        if not numeric_features:
            import pandas as _pd
            return _pd.Series(dtype=float)
        if self._cached_variances is not None:
            cached_cols = set(self._cached_variances.index)
            missing = [f for f in numeric_features if f not in cached_cols]
            if not missing:
                return self._cached_variances[numeric_features]
        self._cached_variances = bulk_variance(df, numeric_features)
        return self._cached_variances

    def _apply_variance_selection(self, df: DataFrame, features: List[str], numeric_set: set) -> None:
        candidates = [f for f in features if f not in self.preserve_features and f in numeric_set]
        if not candidates:
            return
        variances = self._get_variances(df, candidates, numeric_set)
        for feature in candidates:
            variance = variances[feature]
            if isna(variance) or variance < self.variance_threshold:
                if feature in self.selected_features:
                    self.selected_features.remove(feature)
                    self.dropped_features.append(feature)
                    self.drop_reasons[feature] = f"low variance ({variance:.6f})"

    def _apply_correlation_selection(self, df: DataFrame, features: List[str], numeric_set: set) -> None:
        numeric_features = [f for f in features if f in numeric_set and f in self.selected_features]
        if len(numeric_features) < 2:
            return

        pre = self._precomputed_corr_matrix
        if pre is not None and set(numeric_features).issubset(pre.columns):
            corr_matrix = pre.loc[numeric_features, numeric_features].abs()
        else:
            corr_matrix = batched_corr_matrix(
                df, numeric_features, progress_fn=self._progress_fn,
                precomputed_medians=self._precomputed_medians,
                precomputed_non_null=self._precomputed_non_null,
            ).abs()
        upper = corr_matrix.where(np.triu(np.ones(corr_matrix.shape), k=1).astype(bool))
        variances = self._get_variances(df, numeric_features, numeric_set)

        to_drop = set()
        for column in upper.columns:
            correlated = upper.index[upper[column] > self.correlation_threshold].tolist()
            for corr_feature in correlated:
                if corr_feature in self.preserve_features:
                    if column not in self.preserve_features:
                        to_drop.add(column)
                elif column in self.preserve_features:
                    to_drop.add(corr_feature)
                else:
                    if variances[column] >= variances[corr_feature]:
                        to_drop.add(corr_feature)
                    else:
                        to_drop.add(column)

        for feature in to_drop:
            if feature in self.selected_features:
                self.selected_features.remove(feature)
                self.dropped_features.append(feature)
                self.drop_reasons[feature] = f"high correlation (> {self.correlation_threshold})"

    def _apply_l1_selection(self, df: DataFrame, features: List[str], numeric_set: set) -> None:
        if not self.target_column:
            raise ValueError("target_column is required for L1_SELECTION")
        if self.target_column not in df.columns:
            raise ValueError(f"target_column '{self.target_column}' not in DataFrame")

        numeric_features = [f for f in features if f not in self.preserve_features and f in numeric_set]
        if not numeric_features:
            return

        from customer_retention.core.compat import _is_spark_pandas
        if _is_spark_pandas(df):
            from customer_retention.core.compat import as_spark_df
            spark_df = as_spark_df(df[numeric_features + [self.target_column]])
            dropped, reasons, scores = _spark_l1_selection(spark_df, self.target_column, numeric_features, reg_param=1.0 / self.l1_C, elastic_net_param=self.l1_ratio)
            self.importance_scores = scores
            for feature in dropped:
                if feature in self.selected_features:
                    self.selected_features.remove(feature)
                    self.dropped_features.append(feature)
                    self.drop_reasons[feature] = reasons[feature]
            return

        work_df = df[numeric_features + [self.target_column]]
        work_df = work_df.dropna(subset=[self.target_column])
        if len(work_df) < 10:
            raise ValueError(
                f"L1_SELECTION requires at least 10 rows with non-null target; "
                f"got {len(work_df)} after dropping NaN targets"
            )

        X = work_df[numeric_features].fillna(0).to_numpy()
        y = work_df[self.target_column].to_numpy()

        from sklearn.linear_model import LogisticRegression
        from sklearn.preprocessing import StandardScaler

        X_scaled = StandardScaler().fit_transform(X)
        model = LogisticRegression(solver="saga", C=self.l1_C, max_iter=2000, l1_ratio=self.l1_ratio)
        model.fit(X_scaled, y)

        coefs = np.abs(model.coef_)
        max_coefs = coefs.max(axis=0) if coefs.ndim == 2 else coefs

        self.importance_scores = {numeric_features[i]: float(max_coefs[i]) for i in range(len(numeric_features))}
        zero_features = [numeric_features[i] for i in range(len(numeric_features)) if max_coefs[i] == 0.0]
        if len(zero_features) == len(numeric_features):
            return
        for feature in zero_features:
            if feature in self.selected_features:
                self.selected_features.remove(feature)
                self.dropped_features.append(feature)
                self.drop_reasons[feature] = "L1 zero coefficient"

    def get_availability_recommendations(self, availability: Optional["FeatureAvailabilityMetadata"]) -> List[AvailabilityRecommendation]:
        if availability is None:
            return []
        recommendations: List[AvailabilityRecommendation] = []
        problem_columns = availability.new_tracking + availability.retired_tracking + availability.partial_window
        for col in problem_columns:
            feat_info = availability.features.get(col)
            if feat_info is None:
                continue
            recommendations.append(AvailabilityRecommendation(
                column=col,
                issue_type=feat_info.availability_type,
                coverage_pct=feat_info.coverage_pct,
                first_valid_date=feat_info.first_valid_date,
                last_valid_date=feat_info.last_valid_date,
                options=self._build_availability_options(col, feat_info.availability_type, feat_info.first_valid_date, feat_info.last_valid_date, feat_info.coverage_pct),
            ))
        return recommendations

    def _build_availability_options(self, col: str, issue_type: str, first_date: Optional[str], last_date: Optional[str], coverage_pct: float) -> List[Dict[str, Any]]:
        options: List[Dict[str, Any]] = []
        options.append({
            "type": "remove",
            "description": f"Remove '{col}' from feature selection (recommended for most cases)",
            "preserves_data": False,
            "recommended": True,
        })
        options.append({
            "type": "add_indicator",
            "description": f"Create '{col}_available' indicator column to flag valid observations",
            "preserves_data": True,
        })
        if issue_type == "new_tracking":
            options.append({
                "type": "filter_window",
                "description": f"Filter training data to start from {first_date}",
                "preserves_data": True,
            })
            options.append({
                "type": "segment_by_cohort",
                "description": f"Train separate models: pre-{first_date} cohort (without feature) vs post-{first_date} cohort (with feature)",
                "preserves_data": True,
            })
        elif issue_type == "retired":
            options.append({
                "type": "filter_window",
                "description": f"Filter test/scoring data to end at {last_date}",
                "preserves_data": True,
            })
            options.append({
                "type": "segment_by_cohort",
                "description": "Use feature only for historical scoring; train fallback model without it for future predictions",
                "preserves_data": True,
            })
        elif issue_type == "partial_window":
            options.append({
                "type": "filter_window",
                "description": f"Use data only within {first_date} to {last_date}",
                "preserves_data": True,
            })
            options.append({
                "type": "segment_by_availability",
                "description": "Train separate models: one using this feature (within window), one without (outside window)",
                "preserves_data": True,
            })
        if coverage_pct >= 30:
            options.append({
                "type": "impute",
                "description": f"Impute missing values (median/mode) - {coverage_pct:.0f}% coverage may be sufficient",
                "preserves_data": True,
            })
        return options


def _import_spark_ml():
    import pyspark.sql.functions as F  # noqa: N812
    from pyspark.ml.classification import LogisticRegression
    from pyspark.ml.feature import VectorAssembler
    return LogisticRegression, VectorAssembler, F


def _spark_l1_selection(
    spark_df: Any, target_column: str, feature_columns: List[str],
    reg_param: float = 1.0, max_iter: int = 2000, elastic_net_param: float = 1.0,
) -> tuple:
    LR, VectorAssembler, F = _import_spark_ml()

    work_df = spark_df.select([F.col(c).cast("double").alias(c) for c in feature_columns] + [F.col(target_column).cast("double").alias(target_column)])
    work_df = work_df.na.drop(subset=[target_column]).na.fill(0.0, subset=feature_columns)

    stats_exprs = []
    for c in feature_columns:
        stats_exprs.extend([F.mean(c).alias(f"__mean__{c}"), F.stddev(c).alias(f"__std__{c}")])
    stats_row = work_df.agg(*stats_exprs).head()

    scaled_cols = []
    for c in feature_columns:
        mean_val = float(stats_row[f"__mean__{c}"] or 0.0)
        std_val = float(stats_row[f"__std__{c}"] or 1.0)
        if std_val == 0.0:
            std_val = 1.0
        scaled_cols.append(((F.col(c) - F.lit(mean_val)) / F.lit(std_val)).alias(c))
    scaled_cols.append(F.col(target_column))
    work_df = work_df.select(scaled_cols)

    assembler = VectorAssembler(inputCols=feature_columns, outputCol="__scaled__", handleInvalid="keep")
    assembled = assembler.transform(work_df).select("__scaled__", target_column)
    del work_df
    lr = LR(featuresCol="__scaled__", labelCol=target_column, elasticNetParam=elastic_net_param, regParam=reg_param, maxIter=max_iter)
    model = lr.fit(assembled)
    coefs = np.abs(model.coefficients.toArray())
    del model, assembled
    scores: Dict[str, float] = {col: float(coefs[i]) for i, col in enumerate(feature_columns)}
    zero_cols = [col for i, col in enumerate(feature_columns) if coefs[i] == 0.0]
    if len(zero_cols) == len(feature_columns):
        return [], {}, scores
    dropped: List[str] = []
    reasons: Dict[str, str] = {}
    for col in zero_cols:
        dropped.append(col)
        reasons[col] = "L1 zero coefficient"
    return dropped, reasons, scores


def extract_precomputed_stats(findings: Any) -> Dict[str, Any]:
    """Extract medians, non-null counts, and variances from ExplorationFindings.

    Returns dict with keys 'medians', 'non_null', 'variances' — pass as
    ``**extract_precomputed_stats(findings)``-style kwargs or individually
    to ``run_selection_pipeline``.
    """
    medians: Dict[str, float] = {}
    non_null: Dict[str, int] = {}
    variances: Dict[str, float] = {}
    total = getattr(findings, "row_count", 0) or 0
    for name, col in getattr(findings, "columns", {}).items():
        um = getattr(col, "universal_metrics", {})
        tm = getattr(col, "type_metrics", {})
        nc = um.get("null_count", 0)
        if total > 0:
            non_null[name] = total - int(nc)
        std = tm.get("std")
        if std is not None:
            variances[name] = float(std) ** 2
        med = tm.get("median")
        if med is not None:
            medians[name] = float(med)
    import pandas as _pd
    return {
        "precomputed_medians": medians or None,
        "precomputed_non_null": non_null or None,
        "precomputed_variances": _pd.Series(variances) if variances else None,
    }


_L1_SAMPLE_ROWS = 400_000


def _thin_by_temporal_stride(ps_df: Any, temporal_column: str, target_rows: int, log: Any) -> Any:
    import pyspark.sql.functions as F  # noqa: N812

    from customer_retention.core.compat import as_spark_df
    from customer_retention.core.compat.spark_backend import _as_pandas_api

    spark_df = as_spark_df(ps_df)
    total = spark_df.count()
    if total <= target_rows:
        return ps_df

    dates = [row[0] for row in spark_df.select(temporal_column).distinct().orderBy(temporal_column).collect()]
    n_dates = len(dates)
    if n_dates <= 4:
        return ps_df

    rows_per_date = total / n_dates
    target_dates = max(4, int(target_rows / rows_per_date))
    stride = max(1, n_dates // target_dates)
    kept = [dates[i] for i in range(n_dates - 1, -1, -stride)]

    thinned = spark_df.filter(F.col(temporal_column).isin(kept)).localCheckpoint(eager=True)
    log(f"    L1 temporal thinning: {total:,} -> ~{len(kept) * int(rows_per_date):,} rows "
        f"({n_dates} -> {len(kept)} dates, stride {stride})")
    return _as_pandas_api(thinned)


def _format_elapsed(seconds: float) -> str:
    m, s = divmod(int(seconds), 60)
    return f"{m}m{s:02d}s" if m else f"{s}s"


def run_selection_pipeline(
    df: DataFrame, target_column: str,
    variance_threshold: float = 0.01, correlation_threshold: float = 0.95,
    l1_enabled: bool = False, max_features: Optional[int] = None,
    preserve_features: Optional[List[str]] = None,
    progress_fn: Optional[Callable[[str], None]] = None,
    precomputed_corr_matrix: Optional[Any] = None,
    l1_C: float = 1.0, l1_ratio: float = 1.0,
    precomputed_variances: Optional[Any] = None,
    precomputed_medians: Optional[Dict[str, float]] = None,
    precomputed_non_null: Optional[Dict[str, int]] = None,
    temporal_column: Optional[str] = None,
    l1_sample_rows: int = _L1_SAMPLE_ROWS,
) -> FeatureSelectionResult:
    log = progress_fn or (lambda msg: print(msg))
    all_dropped: List[str] = []
    all_reasons: Dict[str, str] = {}
    importance_scores: Optional[Dict[str, float]] = None
    current_df = df
    pipeline_start = time.monotonic()
    n_features_initial = len([c for c in df.columns if c != target_column])

    stages = ["variance", "correlation"]
    if l1_enabled:
        stages.append("L1")
    total_stages = len(stages)

    log(f"Feature selection pipeline: {n_features_initial} features, {total_stages} stages")

    t0 = time.monotonic()
    log(f"  [1/{total_stages}] Variance filter (threshold={variance_threshold})...")
    var_selector = FeatureSelector(
        method=SelectionMethod.VARIANCE, variance_threshold=variance_threshold,
        target_column=target_column, preserve_features=preserve_features,
        precomputed_variances=precomputed_variances,
    )
    result_var = var_selector.fit_transform(current_df)
    # Share cached variances with subsequent stages to avoid recomputation
    shared_variances = var_selector._cached_variances
    current_df = result_var.df
    all_dropped.extend(result_var.dropped_features)
    all_reasons.update(result_var.drop_reasons)
    elapsed = time.monotonic() - t0
    remaining = len(result_var.selected_features)
    log(f"    {_format_elapsed(elapsed)} — dropped {len(result_var.dropped_features)}, remaining {remaining}")

    t0 = time.monotonic()
    log(f"  [2/{total_stages}] Correlation filter (threshold={correlation_threshold}, {remaining} features)...")
    last_method = SelectionMethod.CORRELATION
    result_corr = FeatureSelector(
        method=SelectionMethod.CORRELATION, correlation_threshold=correlation_threshold,
        target_column=target_column, preserve_features=preserve_features,
        precomputed_corr_matrix=precomputed_corr_matrix, progress_fn=log,
        precomputed_variances=shared_variances,
        precomputed_medians=precomputed_medians,
        precomputed_non_null=precomputed_non_null,
    ).fit_transform(current_df)
    current_df = result_corr.df
    all_dropped.extend(result_corr.dropped_features)
    all_reasons.update(result_corr.drop_reasons)
    elapsed = time.monotonic() - t0
    remaining = len(result_corr.selected_features)
    log(f"    {_format_elapsed(elapsed)} — dropped {len(result_corr.dropped_features)}, remaining {remaining}")

    if l1_enabled:
        import gc
        del result_var, result_corr, shared_variances
        gc.collect()

        from customer_retention.core.compat import _is_spark_pandas
        if temporal_column and temporal_column in current_df.columns:
            if _is_spark_pandas(current_df):
                current_df = _thin_by_temporal_stride(current_df, temporal_column, l1_sample_rows, log)
            current_df = current_df.drop(columns=[temporal_column])

        t0 = time.monotonic()
        log(f"  [3/{total_stages}] L1 selection ({remaining} features, max_features={max_features})...")
        last_method = SelectionMethod.L1_SELECTION
        result_l1 = FeatureSelector(
            method=SelectionMethod.L1_SELECTION, target_column=target_column,
            preserve_features=preserve_features, max_features=max_features,
            l1_C=l1_C, l1_ratio=l1_ratio,
        ).fit_transform(current_df)
        current_df = result_l1.df
        all_dropped.extend(result_l1.dropped_features)
        all_reasons.update(result_l1.drop_reasons)
        importance_scores = result_l1.importance_scores
        elapsed = time.monotonic() - t0
        remaining = len(result_l1.selected_features)
        log(f"    {_format_elapsed(elapsed)} — dropped {len(result_l1.dropped_features)}, remaining {remaining}")

    total_elapsed = time.monotonic() - pipeline_start
    selected = [c for c in current_df.columns if c != target_column]
    log(f"  Done: {n_features_initial} -> {len(selected)} features ({len(all_dropped)} dropped) in {_format_elapsed(total_elapsed)}")

    return FeatureSelectionResult(
        df=current_df, selected_features=selected, dropped_features=all_dropped,
        drop_reasons=all_reasons, method_used=last_method, importance_scores=importance_scores,
    )



#!/usr/bin/env python3
"""Generate requirements.txt and requirements-databricks.txt from pyproject.toml.

Run manually:   python scripts/generate_requirements.py
Pre-commit hook: auto-runs on push (see .pre-commit-config.yaml)
"""
from __future__ import annotations

import sys
from pathlib import Path

_PROJECT_ROOT = Path(__file__).resolve().parent.parent
sys.path.insert(0, str(_PROJECT_ROOT / "src"))

from customer_retention.integrations.requirements_generator import write_requirements  # noqa: E402

if __name__ == "__main__":
    req_path, dbr_path = write_requirements(_PROJECT_ROOT)
    print(f"Generated {req_path}")
    print(f"Generated {dbr_path}")
